from __future__ import annotations

import json
import logging
import os
from collections.abc import Callable, Hashable
from typing import Any, ClassVar, Self, cast
from uuid import UUID

import networkx as nx
from aviary.message import Message, join
from aviary.tools import ToolRequestMessage, ToolResponseMessage
from pydantic import BaseModel, ConfigDict, Field, JsonValue, field_validator

from ldp.graph import OpResult
from ldp.utils import discounted_returns

logger = logging.getLogger(__name__)


class Transition(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    # Sentinel value for missing observation, as opposed to empty observation
    # Only used for tests; a user should never use this.
    NO_OBSERVATION: ClassVar[list[Message]] = []

    timestep: int = Field(description="Zero-indexed MDP timestep t.")

    agent_state: Any = Field(
        description=(
            "Agent.get_asv's input. This is `s_t` in RL terms. Note that `s_0` comes"
            " from `Agent.init_state()`"
        )
    )
    next_agent_state: Any = Field(
        description="Agent.get_asv's output. This is s_t+1 in RL terms."
    )

    observation: list[ToolResponseMessage | Message] = Field(
        description="Agent.get_asv's input. This is o_t in RL terms."
    )
    next_observation: list[ToolResponseMessage | Message] = Field(
        description="Environment.step output. This is o_t+1 in RL terms."
    )

    action: OpResult[ToolRequestMessage] | None = Field(
        default=None, description="Agent.get_asv output. This is a_t in RL terms."
    )

    reward: float = Field(
        default=0.0, description="Environment.step output. This is r_t in RL terms."
    )

    truncated: bool = Field(
        default=False, description="timestep t's Environment.step output."
    )
    done: bool = Field(
        default=False, description="timestep t's Environment.step output."
    )
    value: float = Field(
        default=0.0,
        description=(
            "Value estimate output from timestep t's Agent.get_asv. This is v(s_t)"
            " [state value function] or q(s_t, a_t) [state-action value]."
        ),
    )
    # JsonValue so we can serialize
    metadata: dict[str, JsonValue] = Field(default_factory=dict)

    @field_validator("action", mode="before")
    @classmethod
    def construct_action(
        cls, action: OpResult[ToolRequestMessage] | dict | None
    ) -> OpResult[ToolRequestMessage] | None:
        if isinstance(action, dict):
            return OpResult.from_dict(ToolRequestMessage, action)
        return action

    @property
    def failed(self) -> bool:
        """Get if an exception was encountered during rollout, for convenience.

        If True, this transition should not be trained on.
        Failed transitions are for debugging purposes.
        """
        return bool(self.metadata.get("exception"))

    def model_dump_json(self, *, indent: int | None = None, **kwargs) -> str:
        # The kwargs for model_dump are the same as super().model_dump_json,
        # with the exception of indent.
        dump = self.model_dump(**kwargs)
        if self.action is not None:
            dump["action"] = self.action.to_dict()
        return json.dumps(dump, indent=indent)


class Trajectory(BaseModel):
    model_config = ConfigDict(extra="forbid")

    traj_id: str | None = None
    steps: list[Transition] = Field(default_factory=list)

    @property
    def failed(self) -> bool:
        return any(step.failed for step in self.steps)

    @property
    def done(self) -> bool:
        if not self.steps:
            return False
        return self.steps[-1].done

    def to_jsonl(self, filename: str | os.PathLike) -> None:
        with open(filename, "w") as f:
            f.write(json.dumps(self.traj_id) + "\n")
            f.writelines(s.model_dump_json() + "\n" for s in self.steps)

    @classmethod
    def from_jsonl(cls, filename: str | os.PathLike) -> Self:
        with open(filename) as f:
            reader = iter(f)
            traj = cls(traj_id=json.loads(next(reader)))
            for json_line in reader:
                traj.steps.append(Transition(**json.loads(json_line)))
        return traj

    def compute_discounted_returns(self, discount: float = 1.0) -> list[float]:
        """Compute the discounted returns for each step in the trajectory."""
        return discounted_returns(
            rewards=[step.reward for step in self.steps],
            terminated=[step.truncated for step in self.steps],
            discount=discount,
        )


class TransitionTree:
    def __init__(self, root_id: str | UUID):
        """A tree of transitions.

        If A->B is an edge in this tree, then A and B are consecutive
        transitions in an LDP. Any path from the root node to a terminal
        node constitutes a complete LDP.

        A node may be assigned an arbitrary weight, which will be treated
        as a relative probability of sampling that node. For example, if
        A(weight=1) and B(w=2) are both children of the same node,
        then we treat B as twice as likely as A.

        Args:
            root_id: A unique identifier for the root node of the tree.
                All IDs of transitions added to this tree must begin with
                the same root_id.
        """
        self.root_id = str(root_id)

        self.tree = nx.DiGraph()  # the actual tree
        self.rev_tree = nx.DiGraph()  # the same as self.tree, but with reversed edges

        self._add_node(self.root_id, transition=None, weight=1.0)

    def _add_node(
        self, step_id: str, transition: Transition | None, weight: float
    ) -> None:
        self.tree.add_node(step_id, transition=transition, weight=weight)
        self.rev_tree.add_node(step_id)

    def _add_edge(self, parent_step_id: str, child_step_id: str) -> None:
        self.tree.add_edge(parent_step_id, child_step_id)
        self.rev_tree.add_edge(child_step_id, parent_step_id)

    def get_transition(self, step_id: str) -> Transition:
        if step_id == self.root_id:
            raise ValueError("Root node has no transition.")

        return cast(Transition, self.tree.nodes[step_id]["transition"])

    def get_weight(self, step_id: str) -> float:
        return cast(float, self.tree.nodes[step_id]["weight"])

    def add_transition(
        self, step_id: str, step: Transition, weight: float = 1.0
    ) -> None:
        """Add a transition to the tree.

        Args:
            step_id: A unique identifier for the root node of the tree.
                The expected form of the step ID is "{parent step ID}:{step index}".
            step: The transition to add.
            weight: Weight of the transition. Defaults to 1.0.
        """
        root_id, *step_ids = step_id.split(":")
        assert (
            root_id == self.root_id
        ), f"Step ID {step_id} does not start with root ID {self.root_id}"
        assert step_ids, "Step ID cannot be the same as the root ID."
        # TODO: maybe this should be warning?
        assert (
            step_id not in self.tree
        ), f"Step ID {step_id} already exists in the tree."

        self._add_node(step_id, transition=step, weight=weight)

        parent_id = ":".join([root_id, *step_ids[:-1]])
        if parent_id in self.tree:
            self._add_edge(parent_id, step_id)

    def get_trajectories(self) -> list[Trajectory]:
        """Return a list of trajectories.

        Since each path from the root node to a terminal node defines
        a unique trajectory, N(terminal node) trajectories will be returned.
        The trajectory ID will be set to the ID of the terminal step.

        Note that we include failed and truncated trajectories; it is up to the
        caller to decide what to do them.

        Returns:
            All trajectories in this tree.
        """
        trajs = []
        step: Transition | None

        for step_id, step in self.tree.nodes(data="transition"):
            if not step:
                # root node
                continue

            is_terminal = (
                # check terminal conditions in increasing order of expense
                step.done
                or step.truncated
                or step.failed
                or self.tree.out_degree(step_id) == 0
            )

            if not is_terminal:
                continue

            # set the ID to the terminal node, which uniquely identifies the path
            traj = Trajectory(traj_id=step_id)
            # Build the trajectory up from a terminal node
            current_step: Transition | None = step
            current_step_id = step_id

            # Walk backwards towards the root (current_step=None)
            while current_step:
                traj.steps.append(current_step)

                parent_step_id, *extra = list(self.rev_tree.successors(current_step_id))
                assert not extra, f"Expected a single parent, but got {len(extra) + 1}"

                current_step_id = parent_step_id
                current_step = self.tree.nodes[parent_step_id]["transition"]

            # would've added things in reverse order, so fix that here
            traj.steps.sort(key=lambda x: x.timestep)
            trajs.append(traj)

        return trajs

    def assign_mc_value_estimates(self, discount_factor: float = 1.0) -> None:
        """Assign Monte Carlo state-action value estimates to each transition (in-place).

        Args:
            discount_factor: The discount factor to use when computing cumulative
                future rewards.
        """
        for step_id in nx.topological_sort(self.rev_tree):
            step: Transition | None = self.tree.nodes[step_id]["transition"]
            if step is None:
                continue

            if children := list(self.tree.successors(step_id)):
                # V_{t+1}(s') = sum_{a'} p(a'|s') * Q_{t+1}(s', a')
                # Here we assume p(a'|s') is uniform over the sampled actions..
                # TODO: don't make that assumption where a logprob is available
                weights = [self.get_weight(child_id) for child_id in children]
                steps = [self.get_transition(child_id) for child_id in children]
                v_tp1 = sum(
                    w * step.value for w, step in zip(weights, steps, strict=True)
                ) / sum(weights)
            else:
                v_tp1 = 0.0

            # Q_t(s_t, a_t) = r_{t+1} + gamma * V_{t+1}(s_{t+1})
            # (we are assuming the environment is deterministic)
            step.value = step.reward + discount_factor * v_tp1

    def merge_identical_nodes(
        self,
        agent_state_hash_fn: Callable[[Any], Hashable],
        observation_hash_fn: Callable[
            [list[ToolResponseMessage | Message]], Hashable
        ] = join,
        next_observation_hash_fn: Callable[
            [list[ToolResponseMessage | Message]], Hashable
        ] = join,
    ) -> TransitionTree:
        """Merge nodes with identical (state, observation, action)s. Returns a new tree.

        Args:
            agent_state_hash_fn: A function that returns a hashable representation
                of the agent state of a transition.
            observation_hash_fn: A function that returns a hashable representation
                of the observation messages of a transition.
            next_observation_hash_fn: A function that returns a hashable representation
                of the next observation messages of a transition.
        """
        new_tree = TransitionTree(self.root_id)

        # step hash -> step ID
        seen_hash_to_step_id: dict[int, str] = {}
        # old step ID -> new step ID
        node_remap: dict[str, str] = {self.root_id: self.root_id}

        for step_id in nx.topological_sort(self.tree):
            step: Transition | None = self.tree.nodes[step_id]["transition"]
            if step is None:
                continue

            state_hash = agent_state_hash_fn(step.agent_state)

            if step.action is not None:
                tool_request_msg = step.action.value
                # NOTE: we are ignoring tool call ID in the comparison of tool requests.
                # Thus, the tool call ID is excluded from the hash, so we can't just
                # simply call str(tool_request_msg)
                action_str = (tool_request_msg.content or "") + " ".join(
                    str(tc) for tc in tool_request_msg.tool_calls
                )
            else:
                action_str = ""

            step_hash = hash((
                state_hash,
                action_str,
                # (s, a, o): works for deterministic envs
                observation_hash_fn(step.observation),
                # (s, a, o, o'): works for both deterministic and stochastic envs
                next_observation_hash_fn(step.next_observation),
            ))
            step_weight = self.get_weight(step_id)

            if step_hash in seen_hash_to_step_id:  # Seen: merge
                merged_step_id = node_remap[step_id] = seen_hash_to_step_id[step_hash]
                # Not sure if this is the fastest way to do this
                new_tree.tree.nodes[merged_step_id]["weight"] += step_weight
            else:  # Unseen: don't merge
                node_remap[step_id] = seen_hash_to_step_id[step_hash] = step_id
                parent_id = node_remap[":".join(step_id.split(":")[:-1])]

                # manually add transitions, since the step_id substring relationship
                # will be broken
                new_tree._add_node(step_id, transition=step, weight=step_weight)
                new_tree._add_edge(parent_id, step_id)

        return new_tree



import contextlib
import json
from typing import Any

from aviary.message import Message
from aviary.tools import Tool, ToolRequestMessage
from aviary.tools.base import ToolCall

from ldp.graph import IdentityOp, OpResult, compute_graph

from .agent import Agent
from .simple_agent import SimpleAgentState

MISSING_DEFAULT = object()
CLEAR = "CLEAR"  # Clears the current input and allows the user to start over, e.g. if they made a mistake
EXIT = "EXIT"  # Exits the agent by raising a RuntimeError. Makes it possible to interrupt a rollout


class InteractiveAgent(Agent[SimpleAgentState]):
    """An "agent" that provides an interface for human users to interact with environments."""

    def __init__(self):
        self.action_op = IdentityOp[ToolRequestMessage]()

    async def init_state(self, tools: list[Tool]) -> SimpleAgentState:
        print()  # add a newline to flush any progress bars, etc
        print("AVAILABLE TOOLS:\n" + ("-" * 80))
        for tool in tools:
            info = tool.info
            docstring = f"{info.name}("
            arg_docs = []
            for pname, pprops in info.parameters.properties.items():
                docstring += self._get_param_string(pname, pprops) + ", "
                arg_doc = "   " + pname
                if "description" in pprops:
                    arg_doc += ": " + pprops.get("description", "")
                arg_docs.append(arg_doc)

            docstring = docstring.rstrip(", ") + "):\n"
            docstring += "   " + info.description + "\n\n"
            docstring += "\n".join(arg_docs)
            docstring += "\n"
            print(docstring)
        print("-" * 80)
        return SimpleAgentState(tools=tools)

    @compute_graph()
    async def get_asv(  # noqa: C901
        self, agent_state: SimpleAgentState, obs: list[Message]
    ) -> tuple[OpResult[ToolRequestMessage], SimpleAgentState, float]:
        print()  # add a newline to flush any progress bars, etc
        print("OBSERVATIONS:\n" + ("-" * 80))
        for msg in obs:
            print((msg.content or "") + "\n")
        print("-" * 80)

        next_agent_state = agent_state.get_next_state(obs)

        tool: Tool | None = None
        while not tool:
            tool_choice = input(">>> Select tool by name: ")
            if tool_choice == CLEAR:
                continue
            if tool_choice == EXIT:
                raise RuntimeError("User requested to kill the agent.")

            tool = next(
                (t for t in agent_state.tools if t.info.name == tool_choice), None
            )
            if not tool:
                print(
                    f"Tool '{tool_choice}' not found. Please select from the above tools."
                )

        params: dict[str, Any] = {}
        for pname, pprops in tool.info.parameters.properties.items():
            pdefault = pprops.get("default", MISSING_DEFAULT)
            prompt = f">>> Enter parameter ({self._get_param_string(pname, pprops)}): "
            while True:
                value = input(prompt)
                if value == CLEAR:
                    return await self.get_asv(agent_state, obs)  # just start over
                if value == EXIT:
                    raise RuntimeError("User requested to kill the agent.")

                with contextlib.suppress(json.JSONDecodeError):
                    # lets us load ints, etc. Otherwise, assume it's a string
                    value = json.loads(value)
                if not value:
                    if pdefault is MISSING_DEFAULT:
                        print("Parameter is required.")
                        continue

                    value = pdefault

                params[pname] = value
                break

        tool_call = ToolCall.from_tool(tool, **params)
        action = await self.action_op(ToolRequestMessage(tool_calls=[tool_call]))

        next_agent_state.messages = [*next_agent_state.messages, action.value]

        return action, next_agent_state, 0.0

    @staticmethod
    def _get_param_string(pname: str, pprops: dict[str, Any]) -> str:
        pstring = pname
        if ptype := (pprops.get("type") or "Any"):
            pstring += f": {ptype}"

        if (pdefault := pprops.get("default", MISSING_DEFAULT)) is not MISSING_DEFAULT:
            pstring += f" = {pdefault}"
        return pstring



from __future__ import annotations

from typing import Any, Self, cast

from aviary.message import Message
from aviary.tools import Tool, ToolRequestMessage, ToolResponseMessage
from pydantic import BaseModel, ConfigDict, Field

from ldp.graph import ConfigOp, LLMCallOp, OpResult, compute_graph
from ldp.llms import prepend_sys

from . import DefaultLLMModelNames
from .agent import Agent


class SimpleAgentState(BaseModel):
    """Simple bucket for an Agent to access tools and store messages."""

    tools: list[Tool] = Field(default_factory=list)
    messages: list[ToolRequestMessage | ToolResponseMessage | Message] = Field(
        default_factory=list
    )

    def get_next_state(
        self,
        obs: list[Message] | None = None,
        tools: list[Tool] | None = None,
        **kwargs,
    ) -> Self:
        """
        Get the next agent state without mutating the optional prior state.

        Do not mutate self here, just read from it.

        Args:
            obs: Optional observation messages to use in creating the next state.
            tools: Optional list of tools available to the agent. If unspecified, these
                should be pulled from the prior_state.
            kwargs: Additional keyword arguments to pass to this class's constructor.

        Returns:
            The next agent state (which is not an in-place change to self).
        """
        return type(self)(
            tools=tools if tools is not None else self.tools,
            messages=self.messages + (obs or []),
            **kwargs,
        )


class SimpleAgent(BaseModel, Agent[SimpleAgentState]):
    """Simple agent that can pick and invoke tools with a language model.

    It does not have a system prompt because it's meant to be lightweight.
    """

    # Freeze to ensure the only mutation happens in either the agent state (which is
    # passed around) or in the internal Ops
    model_config = ConfigDict(frozen=True)

    llm_model: dict[str, Any] = Field(
        default={"model": DefaultLLMModelNames.OPENAI.value, "temperature": 0.1},
        description="Starting configuration for the LLM model. Trainable.",
    )
    sys_prompt: str | None = Field(
        default=None,
        description=(
            "Opt-in system prompt. If one is passed, the system prompt is not set up to"
            " be trainable, because this class is meant to be quite simple as far as"
            " possible hyperparameters."
        ),
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._config_op = ConfigOp[dict](config=self.llm_model)
        self._llm_call_op = LLMCallOp()

    async def init_state(self, tools: list[Tool]) -> SimpleAgentState:
        return SimpleAgentState(tools=tools)

    @compute_graph()
    async def get_asv(
        self, agent_state: SimpleAgentState, obs: list[Message]
    ) -> tuple[OpResult[ToolRequestMessage], SimpleAgentState, float]:
        next_state = agent_state.get_next_state(obs)

        messages = (
            prepend_sys(next_state.messages, sys_content=self.sys_prompt)
            if self.sys_prompt is not None
            else next_state.messages
        )
        result = cast(
            OpResult[ToolRequestMessage],
            await self._llm_call_op(
                await self._config_op(), msgs=messages, tools=next_state.tools
            ),
        )
        next_state.messages = [*next_state.messages, result.value]
        return result, next_state, 0.0



import inspect
from collections.abc import Mapping, Sequence
from typing import Any, ClassVar

try:
    import torch
    from torch import nn
except ImportError:
    raise ImportError(
        "ldp.graph.torch_ops requires PyTorch as a dependency. "
        "Please run `pip install ldp[nn]`."
    ) from None

from .async_torch import async_protect_torch_call
from .op_utils import CallID, get_call_id, get_training_mode
from .ops import GradInType, Op, OpCtx, ResultOrValue


class TorchOp(Op[torch.Tensor]):
    """An operation that wraps a PyTorch module."""

    CTX_TENSOR_INPUT_KEY: ClassVar[str] = "tensor_input"

    def __init__(self, module: nn.Module):
        super().__init__()
        self.module = module

        # override forward args with the signature of the function
        fwd_sig = inspect.signature(self.module.forward)
        self._fwd_args = list(fwd_sig.parameters.values())

    def __str__(self) -> str:
        return f"{type(self).__name__} {type(self.module).__name__} ({id(self)})"

    async def forward(self, *args, **kwargs: Any) -> torch.Tensor:
        tensor_args = [
            arg
            if isinstance(arg, torch.Tensor)
            else torch.tensor(arg, requires_grad=True)
            for arg in args
        ]
        tensor_kwargs = {
            k: v if isinstance(v, torch.Tensor) else torch.tensor(v, requires_grad=True)
            for k, v in kwargs.items()
        }

        is_training = get_training_mode()

        if is_training:
            store_tensor_inputs(
                ctx=self.ctx,
                key=self.CTX_TENSOR_INPUT_KEY,
                tensor_args=tensor_args,
                tensor_kwargs=tensor_kwargs,
                fwd_args=self._fwd_args,
            )

        return await async_protect_torch_call(self.module, no_grad=not is_training)(
            *tensor_args, **tensor_kwargs
        )

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: float | torch.Tensor,
        call_id: CallID,
    ) -> GradInType:
        tensor_args, tensor_kwargs = ctx.get(call_id, cls.CTX_TENSOR_INPUT_KEY)
        n_pos_args = len(tensor_args)
        output = ctx.get(call_id, "output").value

        if not isinstance(grad_output, torch.Tensor):
            grad_output = torch.tensor(grad_output, dtype=output.dtype)
        grad_output = grad_output.to(output.device)

        while grad_output.ndim < output.ndim:
            # Assume we can broadcast, so expand dims
            # e.g. if output.shape = (2, 1, 1) and grad_output is a scalar
            # then we want to expand to (1, 1, 1) and then broadcast
            grad_output = grad_output.unsqueeze(-1)

        if output.shape != grad_output.shape:
            raise RuntimeError(
                f"Output shape {output.shape} does not match grad_output shape {grad_output.shape}"
            )

        gradients = torch.autograd.grad(
            output,
            [*tensor_args, *tensor_kwargs.values()],
            grad_outputs=grad_output,
            allow_unused=True,
            retain_graph=True,
        )

        grad_args = [grad.detach().cpu().float() for grad in gradients[:n_pos_args]]
        grad_kwargs = {
            k: grad.detach().cpu().float()
            for k, grad in zip(
                tensor_kwargs.keys(), gradients[n_pos_args:], strict=True
            )
        }

        return grad_args, grad_kwargs


def store_tensor_inputs(
    ctx: OpCtx,
    key: str,
    tensor_args: Sequence[torch.Tensor],
    tensor_kwargs: Mapping[str, torch.Tensor],
    fwd_args: Sequence[inspect.Parameter],
    detach: bool = False,
) -> None:
    call_id = get_call_id()
    # Save tensor inputs for backward pass. Do not clobber "input", since
    # that is needed for compute graph. Map positional args to kwargs
    # Copying so we don't modify tensor_kwargs in-place
    ctx_kwargs = tensor_kwargs.copy()  # type: ignore[attr-defined]

    # See Op.__call__ for some notes on what this is doing.
    for i_arg, (arg, param) in enumerate(
        # strict=False b/c not all params in _fwd_args will be in args (i.e. defaults and **kwargs)
        zip(tensor_args, fwd_args, strict=False)
    ):
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            ctx_args = list(tensor_args[i_arg:])
            break

        # Normal positional arg
        ctx_kwargs[param.name] = arg
    else:
        ctx_args = []  # if we got here, there were no *args

    if detach:
        # Detach the tensors from the compute graph and move to CPU
        ctx_args = [arg.detach().cpu() for arg in ctx_args]
        ctx_kwargs = {k: v.detach().cpu() for k, v in ctx_kwargs.items()}

    ctx.update(call_id, key, (ctx_args, ctx_kwargs))



__all__ = ["tqdm", "trange"]

import logging
import os

from tqdm import tqdm, trange

logger = logging.getLogger(__name__)

if os.environ.get("LDP_TQDM_USE_RICH", "").lower() in {"1", "true", "yes"}:
    # TODO: remove after https://github.com/tqdm/tqdm/issues/1618
    try:
        # pylint: disable-next=reimported
        from tqdm.rich import tqdm, trange  # type: ignore[no-redef]
    except ModuleNotFoundError:
        logger.warning(
            "User opted into rich progress via the environment variable"
            " LDP_TQDM_USE_RICH, but did not have 'rich' installed."
            " Please run `pip install ldp[rich]`.",
            exc_info=True,
        )



import os
import secrets
from typing import TYPE_CHECKING, Annotated, TypeVar

import httpx
from aviary.message import Message
from aviary.tools import Messages, Tool, ToolRequestMessage, ToolsAdapter
from pydantic import BaseModel

from ldp.graph import OpResult, get_training_mode

from .agent import Agent
from .simple_agent import SimpleAgentState

if TYPE_CHECKING:
    from fastapi import FastAPI

TSerializableAgentState = TypeVar("TSerializableAgentState", bound=BaseModel)


class HTTPAgentClient(Agent[TSerializableAgentState]):
    """Interact with an Agent running in a server via POST requests."""

    def __init__(
        self,
        agent_state_type: type[TSerializableAgentState],
        server_url: str,
        request_headers: httpx._types.HeaderTypes | None = None,
        request_timeout: float | None = None,
    ):
        super().__init__()
        self._agent_state_type = agent_state_type
        self._request_url = server_url
        self._request_headers = request_headers
        self._request_timeout = request_timeout

    async def get_asv(
        self,
        agent_state: TSerializableAgentState,
        obs: list[Message],
    ) -> tuple[OpResult[ToolRequestMessage], TSerializableAgentState, float]:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                f"{self._request_url}/get_asv",
                json={
                    "agent_state": agent_state.model_dump(),
                    "obs": [m.model_dump() for m in obs],
                    "training": get_training_mode(),
                },
                headers=self._request_headers,
                timeout=self._request_timeout,
            )
            response.raise_for_status()
            response_data = response.json()
            return (
                OpResult.from_dict(ToolRequestMessage, response_data[0]),
                self._agent_state_type(**response_data[1]),
                response_data[2],
            )

    async def init_state(self, tools: list[Tool]) -> TSerializableAgentState:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                f"{self._request_url}/init_state",
                json=ToolsAdapter.dump_python(tools),
                headers=self._request_headers,
                timeout=self._request_timeout,
            )
            response.raise_for_status()
            return self._agent_state_type(**response.json())


def make_simple_agent_server(
    agent: Agent[SimpleAgentState], render_docs: bool = False
) -> "FastAPI":
    """
    Make a FastAPI app designed to work with the above HTTPAgentClient.

    Here's how this works:
    1. There is an entity orchestrating an Agent's interactions with an Environment.
       A simple example of this is an integration test that sequentially calls
       Agent.get_asv and Environment.step.
    2. That entity is given the above HTTPAgentClient. Any Agent.init_state or
       Agent.get_asv calls the orchestration entity makes are actually
       POST requests under the hood. The agent's "brains" aren't local.
    3. Remotely, this server code is running, and is where the actual Agent logic lives.
       An example of this is a remote server containing GPU(s).
    """
    try:
        from fastapi import Body, Depends, FastAPI, HTTPException, status
        from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
    except ModuleNotFoundError as exc:
        raise ImportError(
            "Please install aviary with the 'server' extra like so:"
            " `pip install aviary[server]`."
        ) from exc

    asgi_app = FastAPI(
        title=f"aviary.Agent {type(agent).__name__}",
        description="Serve inference endpoints for an aviary.Agent with a SimpleAgentState",
        # Only render Swagger docs if local since we don't have a login here
        docs_url="/docs" if render_docs else None,
        redoc_url="/redoc" if render_docs else None,
    )
    auth_scheme = HTTPBearer()

    async def validate_token(
        token: Annotated[HTTPAuthorizationCredentials, Depends(auth_scheme)],
    ) -> HTTPAuthorizationCredentials:
        # NOTE: don't use os.environ.get() to avoid possible empty string matches, and
        # to have clearer server failures if the AUTH_TOKEN env var isn't present
        if not secrets.compare_digest(token.credentials, os.environ["AUTH_TOKEN"]):
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Incorrect bearer token",
                headers={"WWW-Authenticate": "Bearer"},
            )
        return token

    @asgi_app.get("/info")
    def info(
        _: Annotated[HTTPAuthorizationCredentials, Depends(validate_token)],
    ) -> dict[str, str]:
        """Get agent metadata, useful for debugging."""
        return {"agent_type": type(agent).__name__}

    @asgi_app.post("/get_asv")
    async def get_asv(
        agent_state: SimpleAgentState,
        obs: Messages,
        _: Annotated[HTTPAuthorizationCredentials, Depends(validate_token)],
        training: Annotated[bool, Body()] = True,
    ) -> tuple[dict, SimpleAgentState, float]:
        if training:
            raise NotImplementedError("Training is not yet supported.")
        action, agent_state, vhat = await agent.get_asv(agent_state, obs)
        return action.to_dict(), agent_state, vhat

    @asgi_app.post("/init_state")
    async def init_state(
        tools: list[Tool],
        _: Annotated[HTTPAuthorizationCredentials, Depends(validate_token)],
    ) -> SimpleAgentState:
        return await agent.init_state(tools)

    return asgi_app



from typing import Any, cast

from pydantic import BaseModel, ConfigDict, Field

from ldp.agent import Agent, MemoryAgent, ReActAgent
from ldp.alg.optimizer.ape import APEOpt, APEScoreFn
from ldp.alg.optimizer.memory import MemoryFactory, MemoryOpt, PositiveMemoryOpt
from ldp.alg.optimizer.opt import _OPTIMIZER_REGISTRY, ChainedOptimizer, Optimizer

_DEFAULT_OPTIMIZER_ERROR_MSG = (
    "Didn't yet implement an optimizer of type {opt_type} for {agent_type}."
)
AUTOSELECT_BEST_OPTIMIZER = None


class OptimizerConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    optimizer_type: str | None = AUTOSELECT_BEST_OPTIMIZER
    optimizer_kwargs: dict[str, Any] = Field(default_factory=dict)


_DEFAULT_OPTIMIZER_MAP: dict[type[Agent], type[Optimizer]] = {
    MemoryAgent: MemoryOpt,
    ReActAgent: APEOpt,
}


def default_optimizer_factory(
    agent: Agent,
    optimizer_cls: str | type[Optimizer] | None = AUTOSELECT_BEST_OPTIMIZER,
    **optimizer_kwargs,
) -> Optimizer:
    """A method that constructs a default optimizer for commonly-used agents.

    Args:
        agent: Agent to construct the optimizer for.
        optimizer_cls: The optimizer class to use. If not specified, we will try a
            default based on the provided agent.
        optimizer_kwargs: Arguments forwarded to optimizer_cls.

    Returns:
        Instantiated optimizer.
    """
    if isinstance(optimizer_cls, str):
        try:
            optimizer_cls = _OPTIMIZER_REGISTRY[optimizer_cls]
        except KeyError:
            raise TypeError(
                f"Optimizer class not supported by default_optimizer_factory: {optimizer_cls}"
            ) from None

    if optimizer_cls is None:
        optimizer_cls = _DEFAULT_OPTIMIZER_MAP.get(agent.__class__)

    # convince mypy that optimizer_cls is a type from here on
    optimizer_cls = cast(type, optimizer_cls)

    if isinstance(agent, MemoryAgent):
        if optimizer_cls != MemoryOpt:
            raise NotImplementedError(
                _DEFAULT_OPTIMIZER_ERROR_MSG.format(
                    opt_type=optimizer_cls.__name__, agent_type=MemoryAgent.__name__
                )
            )
        return MemoryOpt.from_agent(agent, **optimizer_kwargs)
    if isinstance(agent, ReActAgent):
        if optimizer_cls != APEOpt:
            raise NotImplementedError(
                _DEFAULT_OPTIMIZER_ERROR_MSG.format(
                    opt_type=optimizer_cls.__name__, agent_type=ReActAgent.__name__
                )
            )
        return APEOpt.from_agent(agent, **optimizer_kwargs)
    raise TypeError(f"Unsupported agent type: {agent.__class__.__name__}")


__all__ = [
    "APEOpt",
    "APEScoreFn",
    "ChainedOptimizer",
    "MemoryFactory",
    "MemoryOpt",
    "Optimizer",
    "OptimizerConfig",
    "PositiveMemoryOpt",
    "default_optimizer_factory",
]



import asyncio
import random
from uuid import UUID

import pytest

from ldp.graph import (
    CallID,
    compute_graph,
    eval_mode,
    get_call_id,
    get_run_id,
    get_training_mode,
    op_call,
    set_training_mode,
    train_mode,
)
from ldp.graph.op_utils import _RUN_ID


@pytest.mark.asyncio
async def test_run_ids():
    # check that we don't create a new run ID if we're
    # already in a context
    async with compute_graph():
        run_id_1 = get_run_id()
        async with compute_graph():
            run_id_2 = get_run_id()
            assert (
                run_id_1 == run_id_2
            ), "Should not create a new run ID if already in a run context."

    # Check that after exiting the context, _RUN_ID is no longer set
    with pytest.raises(LookupError):
        _RUN_ID.get()

    # Now check that we don't clobber in coroutines
    async def run_id_test() -> UUID:
        async with compute_graph():
            # Wait randomly, giving time for other calls to potentially clobber
            # this one.
            await asyncio.sleep(random.uniform(0.1, 0.5))
            return get_run_id()

    run_ids = await asyncio.gather(*[run_id_test() for _ in range(10)])
    assert len(run_ids) == len(
        set(run_ids)
    ), "At least two compute graphs had the same run ID, which indicates a clobber."


@pytest.mark.asyncio
async def test_call_ids():
    # check that we don't create a new call ID if we're
    # already in a context
    async with compute_graph(), op_call():
        call_id_1 = get_call_id()
        async with op_call():
            call_id_2 = get_call_id()
            assert (
                call_id_1 == call_id_2
            ), "Should not create a new call ID if already in a call context."

    # Check we cannot create a call ID if not in a run context
    with pytest.raises(RuntimeError, match=r".*not inside compute graph context.*"):
        async with op_call():
            pass

    async def call_test() -> CallID:
        async with op_call():
            # Wait randomly, giving time for other calls to potentially clobber
            # this one.
            await asyncio.sleep(random.uniform(0.1, 0.5))
            return get_call_id()

    async with compute_graph():
        call_ids = await asyncio.gather(*[call_test() for _ in range(10)])
    assert len(call_ids) == len(
        set(call_ids)
    ), "At least two compute graphs had the same run ID, which indicates a clobber."


@pytest.mark.asyncio
async def test_training_mode():
    with eval_mode():
        assert not get_training_mode(), "Training mode was not set to False"
    assert (
        get_training_mode()
    ), "Training mode should have been reset to True after exiting context"

    set_training_mode(False)
    with train_mode():
        assert get_training_mode(), "Training mode was not set to True"
    assert (
        not get_training_mode()
    ), "Training mode should have been reset to False after exiting context"

    # Put back to training for next round of tests
    set_training_mode(True)

    async def training_mode_test(i: int):
        train = i % 2 == 0
        set_training_mode(train)
        # wait a random amount of time to give a chance for other coroutines to clobber
        await asyncio.sleep(random.uniform(0.1, 0.5))

        assert get_training_mode() == train, "Training mode was overwritten."

    await asyncio.gather(*[training_mode_test(i) for i in range(10)])
    # Make sure training_mode_test didn't change our training mode
    assert get_training_mode()

    # Make sure nesting works
    with train_mode():
        assert get_training_mode()
        with eval_mode():
            assert not get_training_mode()
            with train_mode():
                assert get_training_mode()
            assert not get_training_mode()
        assert get_training_mode()

    # Make sure modes are set correctly in coroutines

    @train_mode()
    async def train_then_eval():
        assert get_training_mode()
        await asyncio.sleep(0.0)
        async with eval_mode():
            assert not get_training_mode()

    @eval_mode()
    async def eval_then_train():
        assert not get_training_mode()
        await asyncio.sleep(0.0)
        async with train_mode():
            assert get_training_mode()

    await asyncio.gather(train_then_eval(), eval_then_train())



import textwrap

from ldp.llms.prompts import indent_xml


def test_indent_xml():
    xml = "<root><child1>foo<child2>line1\nline2\nline3</child2></child1></root>"
    expected = textwrap.dedent(
        """\
        <root>
          <child1>
            foo
            <child2>
              line1
              line2
              line3
            </child2>
          </child1>
        </root>"""
    )
    assert indent_xml(xml) == expected



"""This module defines the MemoryAgent class, which extends a base agent model with memory.

capabilities. The MemoryAgent can pick and invoke tools based on the stored and retrieved
memories, formatted using specified prompts. A memory is typically a set of previous trajectories
"""

from collections.abc import Awaitable, Callable, Iterable
from typing import ClassVar, cast

from aviary.message import Message
from aviary.tools import ToolRequestMessage
from pydantic import ConfigDict, Field

from ldp.graph import (
    FxnOp,
    Memory,
    MemoryModel,
    MemoryOp,
    OpResult,
    PromptOp,
    compute_graph,
)
from ldp.llms.prompts import indent_xml

from .simple_agent import SimpleAgent, SimpleAgentState


async def _default_query_factory(messages: Iterable[Message]) -> str:
    return "\n\n".join([str(m) for m in messages if m.role != "system"])


class MemoryAgent(SimpleAgent):
    """
    Simple agent that can pick and invoke tools with memory.

    NOTE: the MemoryAgent does not maintain an explicit value estimate,
    it simply supplies previous trajectories via the prompt.
    As such, the value estimate vhat will always be zero.
    """

    # Working around https://github.com/pydantic/pydantic/issues/10551
    default_query_factory: ClassVar[Callable[[Iterable[Message]], Awaitable[str]]] = (
        _default_query_factory
    )

    prompt: str = Field(
        default=(
            "<episode-memories>\n<description>\n"
            "These are relevant memories from previous attempts at similar tasks, "
            "along with the action taken and the discounted cumulative reward from that action. "
            "A negative reward is failure, a positive reward is success.\n"
            "</description>{memories}</episode-memories>\n\n"
            "Considering the memories, choose the next action."
        ),
        description="Prompt that includes the memories.",
    )
    query_factory: Callable[[Iterable[Message]], Awaitable[str]] = Field(
        default=default_query_factory,
        description=(
            "Async function to generate a Memory query string from messages. It's async"
            " so this can involve an LLM completion if desired."
        ),
        exclude=True,
    )
    memory_prompt: str = Field(
        default="<memory><obs>{input}</obs><action>{output}</action><reward>{value}</reward></memory>",
        description="Prompt for formatting an individual memory. "
        "Use XML instead of JSON to avoid potential escaping issues.",
    )
    num_memories: int = Field(
        default=MemoryModel.DEFAULT_MEMORY_MATCHES,
        description="Number of memories to retrieve from MemoryOp",
    )
    # Freeze to ensure the only mutation happens in either the agent state (which is
    # passed around) or in the internal Ops
    model_config = ConfigDict(frozen=True)

    @staticmethod
    def _format_memories(prompt: str, memories: Iterable[Memory]) -> str:
        return indent_xml(
            "\n".join([
                prompt.format(**m.model_dump(exclude={"run_id", "template"}))
                for m in memories
            ])
        )

    @staticmethod
    def _package_messages(
        msgs: list[Message], memory_prompt: str, use_memories: bool
    ) -> list[Message]:
        if use_memories:
            return [*msgs, Message(content=memory_prompt)]
        return msgs

    def __init__(self, memory_model: MemoryModel | None = None, **kwargs):
        super().__init__(**kwargs)
        self._query_factory_op = FxnOp[str](self.query_factory)
        self._memory_op = MemoryOp(memory_model)
        self._format_memory_op = FxnOp(self._format_memories)
        self._prompt_op = PromptOp(self.prompt)
        self._package_op = FxnOp(self._package_messages)

    @compute_graph()
    async def get_asv(
        self, agent_state: SimpleAgentState, obs: list[Message]
    ) -> tuple[OpResult[ToolRequestMessage], SimpleAgentState, float]:
        next_state = agent_state.get_next_state(obs)

        memories = await self._memory_op(
            query=await self._query_factory_op(next_state.messages),
            matches=self.num_memories,
        )
        packaged_messages = await self._package_op(
            next_state.messages,
            memory_prompt=await self._prompt_op(
                memories=await self._format_memory_op(self.memory_prompt, memories)
            ),
            use_memories=bool(memories.value),
        )
        result = cast(
            OpResult[ToolRequestMessage],
            await self._llm_call_op(
                await self._config_op(), msgs=packaged_messages, tools=next_state.tools
            ),
        )
        next_state.messages = [*next_state.messages, result.value]
        return result, next_state, 0.0



import pytest

from ldp.graph import Memory
from ldp.graph.memory import UIndexMemoryModel
from ldp.llms import EmbeddingModel


@pytest.fixture(name="sample_memory")
def fixture_sample_memory() -> Memory:
    return Memory(
        query="sample string representation", output="observation", value=42.0
    )


@pytest.mark.asyncio
async def test_memory_model_initialization() -> None:
    model = UIndexMemoryModel()
    assert model.embedding_model is not None
    model.model_dump()  # Check we can serialize

    model_custom = UIndexMemoryModel(
        embedding_model=EmbeddingModel.from_name("text-embedding-3-small")
    )
    model_custom.model_dump()  # Check we can serialize


@pytest.mark.asyncio
async def test_add_then_get_memory(sample_memory: Memory) -> None:
    memory_model = UIndexMemoryModel(
        embedding_model=EmbeddingModel.from_name("text-embedding-3-small")
    )
    async with memory_model.safe_access_index() as index:
        assert len(index) == 0, "Should have no memories"
    await memory_model.add_memory(sample_memory)
    async with memory_model.safe_access_index() as index:
        assert len(index) == 1, "Should have one memory"
    assert memory_model.memories[0] == sample_memory
    result = await memory_model.get_memory("sample query", matches=1)
    assert len(result) == 1
    assert result[0] == sample_memory


@pytest.mark.asyncio
async def test_initialization_without_embedding():
    model = UIndexMemoryModel()
    assert isinstance(
        model.embedding_model, EmbeddingModel
    ), "Default embedding model should be set"



import networkx as nx
import pytest
from aviary.message import Message, join
from aviary.tools import ToolResponseMessage

from ldp.data_structures import Transition, TransitionTree


def test_tree_mc_value():
    root_id = "dummy"
    tree = TransitionTree(root_id=root_id)

    kw = {
        "agent_state": None,
        "next_agent_state": None,
        "observation": Transition.NO_OBSERVATION,
        "next_observation": Transition.NO_OBSERVATION,
        "action": None,
    }

    # Construct a tree with some rewards scattered about
    tree.add_transition(f"{root_id}:0", Transition(timestep=0, reward=0.0, **kw))

    tree.add_transition(f"{root_id}:0:0", Transition(timestep=1, reward=1.0, **kw))
    for i in range(3):
        tree.add_transition(
            f"{root_id}:0:0:{i}",
            Transition(timestep=2, reward=float(i), done=True, **kw),
        )

    tree.add_transition(
        f"{root_id}:0:1", Transition(timestep=1, reward=-1.0, done=True, **kw)
    )

    tree.assign_mc_value_estimates(discount_factor=0.9)

    # Now make sure the value estimates are as expected
    # First, check the terminal nodes: Q==reward
    for i in range(3):
        assert tree.get_transition(f"{root_id}:0:0:{i}").value == float(i)
    assert tree.get_transition(f"{root_id}:0:1").value == -1.0

    # Then go up the tree
    assert tree.get_transition(f"{root_id}:0:0").value == pytest.approx(
        1 + 0.9 * ((0 + 1 + 2) / 3), rel=0.001
    )
    assert tree.get_transition(f"{root_id}:0").value == pytest.approx(
        0.0 + 0.9 * ((1.9 - 1) / 2), rel=0.001
    )


def test_tree_node_merging() -> None:
    root_id = "dummy"
    tree = TransitionTree(root_id=root_id)

    kw = {
        "next_agent_state": None,
        "observation": Transition.NO_OBSERVATION,
        "next_observation": Transition.NO_OBSERVATION,
        "action": None,
    }

    # Construct a tree with two identical nodes
    for child_id in ("0", "1"):
        tree.add_transition(
            f"{root_id}:{child_id}",
            Transition(timestep=0, reward=0.0, agent_state=0, **kw),
        )
    # Add some identical nodes from a stochastic environment
    for child_id, next_o_content in zip(("2", "3"), ("stub", "stub 123"), strict=True):
        tree.add_transition(
            f"{root_id}:{child_id}",
            Transition(
                timestep=0,
                reward=0.0,
                agent_state=0,
                **(kw | {"next_observation": [Message(content=next_o_content)]}),
            ),
        )

    # Now some children nodes
    for parent in ("0", "1"):
        tree.add_transition(  # This tests merge of child nodes too
            f"{root_id}:{parent}:0",
            Transition(timestep=1, reward=0.0, agent_state=1, **kw),
        )
    for parent in ("2", "3"):
        tree.add_transition(
            f"{root_id}:{parent}:0",
            Transition(timestep=1, reward=0.0, agent_state=9, **kw),
        )

    # Tree at this stage is:
    # ROOT -> 0 -> 0:0; ROOT -> 1 -> 1:0; ROOT -> 2 -> 2:0; ROOT -> 3 -> 3:0

    def custom_next_obs_join(obs: list[ToolResponseMessage | Message]) -> str:
        return join(
            type(o)(
                content=(o.content or "").split(" ", maxsplit=1)[0],
                **o.model_dump(exclude={"content"}),
            )
            for o in obs
        )

    merged_tree = tree.merge_identical_nodes(
        lambda state: state, next_observation_hash_fn=custom_next_obs_join
    )
    # Tree should now be ROOT -> 0/1 -> 0:0/1:0; ROOT -> 2/3 -> 2:0

    assert {
        step_id: tree.get_weight(step_id) for step_id in nx.topological_sort(tree.tree)
    } == {
        "dummy": 1.0,  # Root
        "dummy:0": 1.0,
        "dummy:1": 1.0,
        "dummy:2": 1.0,
        "dummy:3": 1.0,
        "dummy:0:0": 1.0,
        "dummy:1:0": 1.0,
        "dummy:2:0": 1.0,
        "dummy:3:0": 1.0,
    }
    assert {
        step_id: merged_tree.get_weight(step_id)
        for step_id in nx.topological_sort(merged_tree.tree)
    } == {
        "dummy": 1.0,  # Root
        "dummy:0": 2.0,  # Middle 0/1
        "dummy:2": 2.0,  # Middle 2/3
        "dummy:0:0": 2.0,  # Bottom 0 (from 0/1)
        "dummy:2:0": 2.0,  # Bottom 0 (from 2/3)
    }



import asyncio
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Self, TypeVar, cast
from uuid import UUID

import numpy as np
import numpy.typing as npt
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    JsonValue,
    PrivateAttr,
    field_validator,
    model_validator,
)
from usearch.index import Index

from ldp.llms import EmbeddingModel

if TYPE_CHECKING:
    from .common_ops import MemoryOp
    from .op_utils import CallID
    from .ops import Op, OpResult, TOutput


class Memory(BaseModel):
    """A single memory about an input, output, and value tuple.

    A memory is a record of an input, output, and resulting value. Typically used
    for prompting a language model. Or, it could be about a whole forward pass where
    the input is the observation and the output is the action taken.
    The query is optionally different and used for
    retrieving the memory. For example, it could be much larger because it won't
    be formatted in the resulting prompt.
    """

    query: str = Field(
        description="String to be embedded into a retrieval key for a memory index."
    )
    input: str | None = Field(
        default=None,
        description=(
            "Some input (e.g. prompt to LLM, observation). If None (default), the input"
            " is set to match the query."
        ),
    )
    output: str = Field(description="Some output (e.g. tool selection).")
    value: float | str = Field(
        description="Measure of the output's quality (e.g. loss)."
    )
    metadata: JsonValue = Field(
        default_factory=dict,
        description=(
            "Optional JSON metadata to store with the memory. An example is storing"
            " information an optimizer can use at training time."
        ),
    )
    run_id: UUID | None = Field(
        default=None,
        description=(
            "Associate run_id for debugging purposes to trace "
            "which forward pass generated the memory."
        ),
    )
    template: str = "Input: {input}\nOutput: {output}\nValue: {value}"

    @model_validator(mode="before")
    @classmethod
    def ensure_query(cls, data: Any) -> Any:
        """Copy input to match the query if input is None."""
        if isinstance(data, dict) and data.get("input") is None:
            data["input"] = data["query"]
        return data

    def __str__(self) -> str:
        return self.template.format(**self.model_dump())

    @classmethod
    def from_ops(
        cls,
        mem_op: "MemoryOp",
        mem_call_id: "CallID",
        output_op: "Op[TOutput]",
        output_call_id: "CallID",
        value: float,
        **kwargs,
    ) -> Self:
        """Create from a MemoryOp, output Op, and their call IDs."""
        query: str = mem_op.ctx.get(mem_call_id, "query")
        memory_input: str | None = mem_op.ctx.get(mem_call_id, "memory_input")
        output_result: OpResult[TOutput] = output_op.ctx.get(output_call_id, "output")
        return cls(
            query=query,
            input=memory_input if memory_input is not None else query,
            output=str(output_result.value),
            value=value,
            run_id=output_call_id.run_id,
            **kwargs,
        )


TIndex = TypeVar("TIndex")


class MemoryModel(BaseModel, Generic[TIndex], ABC):
    """A collection of memories with retrieval."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    embedding_model: EmbeddingModel = Field(
        default_factory=lambda: EmbeddingModel.from_name(
            "hybrid-text-embedding-3-small"
        )
    )
    memories: dict[int, Memory] = Field(default_factory=dict)
    _index: TIndex
    _index_lock: asyncio.Lock = PrivateAttr(default_factory=asyncio.Lock)

    @field_validator("memories")
    @classmethod
    def enforce_empty(cls, v: dict) -> dict:
        if v:
            raise ValueError("Memories must be empty at construction time.")
        return v

    async def add_memory(self, memory: Memory) -> None:
        key = await self._add_to_index(
            embedding=await self.embedding_model.embed_text(memory.query)
        )
        self.memories[key] = memory

    DEFAULT_MEMORY_MATCHES: ClassVar[int] = 3

    async def get_memory(
        self, query: str, matches: int = DEFAULT_MEMORY_MATCHES
    ) -> list[Memory]:
        return await self._search_index(
            embedding=await self.embedding_model.embed_text(query), matches=matches
        )

    def __len__(self) -> int:
        return len(self.memories)

    @asynccontextmanager
    async def safe_access_index(self) -> AsyncIterator[TIndex]:
        """Get the internal Index under the protection of an internal Lock."""
        # pylint bug, SEE: https://github.com/pylint-dev/pylint/issues/9813
        async with self._index_lock:  # pylint: disable=not-async-context-manager
            yield self._index

    @abstractmethod
    async def _add_to_index(self, embedding: np.ndarray) -> int:
        """Add an embedding to the internal Index and return its key."""

    @abstractmethod
    async def _search_index(
        self, embedding: np.ndarray, matches: int = DEFAULT_MEMORY_MATCHES
    ) -> list[Memory]:
        """Search the internal Index, returning a 'matches' amount of Memories."""


class UIndexMemoryModel(MemoryModel[Index]):
    """Memory model using a U-Search index."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.embedding_model.dimensions:
            raise TypeError("Specify dimensions to the embedding model.")
        self._index = Index(ndim=self.embedding_model.dimensions)

    async def _add_to_index(self, embedding: np.ndarray) -> int:
        async with self.safe_access_index() as index:
            added_value = cast(
                npt.NDArray[np.int_], index.add(len(self.memories), embedding)
            )
            return added_value.item()

    async def _search_index(
        self, embedding: np.ndarray, matches: int = MemoryModel.DEFAULT_MEMORY_MATCHES
    ) -> list[Memory]:
        async with self.safe_access_index() as index:
            search_matches = index.search(embedding, matches)
        # mypy doesn't respect "old style" __getitem__/__len__ as iterable,
        # so we have this ignore. SEE: https://github.com/python/mypy/issues/9737
        return [self.memories[m.key] for m in search_matches]  # type: ignore[union-attr]



from importlib.metadata import version
from typing import ClassVar

import litellm
import pytest
from aviary.env import DummyEnv, DummyEnvState
from aviary.message import Message
from aviary.tools import Tool, ToolCall, ToolRequestMessage, ToolResponseMessage

from ldp.agent import SimpleAgent

from . import CILLMModelNames


class ParallelizedDummyEnv(DummyEnv):
    def __init__(self, right_hand_broken: bool = False):
        super().__init__()
        self.right_hand_broken = right_hand_broken

    RIGHT_HAND_BROKEN_MESSAGE: ClassVar[str] = "Right hand is broken."

    async def reset(self) -> tuple[list[Message], list[Tool]]:
        def move_right_hand(
            distance: int,  # noqa: ARG001
            state: DummyEnvState,
        ) -> None:
            """
            Move your right hand forward or backward.

            Args:
                distance: Integer distance to move (mm), where forward is positive.
                state: Current state.
            """
            if self.right_hand_broken:  # Use this to test tool errors
                raise RuntimeError(self.RIGHT_HAND_BROKEN_MESSAGE)
            state.reward += 1

        def move_left_hand(
            distance: int,  # noqa: ARG001
            state: DummyEnvState,
        ) -> None:
            """
            Move your left hand forward or backward.

            Args:
                distance: Integer distance to move (mm), where forward is positive.
                state: Current state.
            """
            state.reward += 1

        def smile_and_wave(state: DummyEnvState) -> None:
            """
            Smile and wave.

            Args:
                state: Current state.
            """
            state.reward = 10
            state.done = True

        self.tools = [
            Tool.from_function(move_left_hand),
            Tool.from_function(move_right_hand),
            Tool.from_function(smile_and_wave),
        ]
        self.state = type(self).State(
            messages=[
                Message(
                    role="user",
                    content=(
                        "You are the president of the United States of America."
                        " Please move both hands at the same time, and then smile"
                        " and wave."
                    ),
                )
            ]
        )
        return self.state.messages, self.tools


class TestParallelism:
    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_SimpleAgent_can_parallel_call(self) -> None:
        env = ParallelizedDummyEnv()
        obs, tools = await env.reset()
        agent = SimpleAgent()

        # Check parallel tool calls
        action, agent_state, _ = await agent.get_asv(
            await agent.init_state(tools=tools), obs
        )
        selected_tools: set[str] = {tc.function.name for tc in action.value.tool_calls}
        assert (
            {"move_left_hand", "move_right_hand"} <= selected_tools
        ), f"Agent should've chosen tools in parallel, but it chose {selected_tools}"

    @pytest.mark.parametrize(
        "model_name", [CILLMModelNames.ANTHROPIC.value, "gpt-4-turbo"]
    )
    @pytest.mark.asyncio
    async def test_exec_tool_calls_handling(self, model_name: str) -> None:
        env = ParallelizedDummyEnv(right_hand_broken=True)
        obs, tools = await env.reset()
        right_hand_tool = tools[1]
        agent = SimpleAgent(
            llm_model=SimpleAgent.model_fields["llm_model"].default
            | {"model": model_name}
        )
        agent_state = await agent.init_state(tools=tools)

        # 1. Let's DIY create a ToolRequestMessage for test determinism
        request_msg = ToolRequestMessage(
            content="stub", tool_calls=[ToolCall.from_tool(right_hand_tool, distance=5)]
        )
        agent_state.messages.extend([*obs, request_msg])

        # 2. Okay, our hand was broken, let's handle it DIY-style
        try:
            obs, *_ = await env.step(action=request_msg)
        except RuntimeError as exc:
            obs = [
                Message(
                    content=f"Failed to execute tools with message:\n{exc}", role="tool"
                )
            ]
        else:
            raise AssertionError("Should have blown up per the test logic.")

        # 2. Well, it looks like both Anthropic and OpenAI don't like DIY-style
        #    (using a bare Message) because they expect a tool call ID and tool name
        with pytest.raises(
            litellm.BadRequestError,
            match=(
                "invalid" if version(litellm.__name__) < "1.45.0" else "tool_call_id"
            ),
        ):
            await agent.get_asv(agent_state, obs)

        # 3. Alright, let's check the agent doesn't blow up if we use a
        #    ToolResponseMessage as Anthropic and OpenAI expect
        await agent.get_asv(
            agent_state,
            ToolResponseMessage.from_request(request_msg, contents=["null"]),  # type: ignore[arg-type]
        )

        # 4. Now that we have confirmed that, let's make sure exec_tool_calls
        #    can automate this for us
        obs = await env.exec_tool_calls(  # type: ignore[assignment]
            message=request_msg, state=env.state, handle_tool_exc=True
        )
        (failure_tool_response,) = obs
        assert isinstance(failure_tool_response, ToolResponseMessage)
        assert env.RIGHT_HAND_BROKEN_MESSAGE in failure_tool_response.content
        await agent.get_asv(agent_state, obs)



"""
A module is a subgraph of a compute graph that can be exposed like a single node/op.

An analogous entity in PyTorch is torch.nn.Module.
"""

from .llm_call import ParsedLLMCallModule
from .react import (
    ReActModule,
    ToolDescriptionMethods,
    parse_message,
)
from .reflect import ReflectModule, ReflectModuleConfig
from .thought import ThoughtModule

__all__ = [
    "ParsedLLMCallModule",
    "ReActModule",
    "ReflectModule",
    "ReflectModuleConfig",
    "ThoughtModule",
    "ToolDescriptionMethods",
    "parse_message",
]



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from collections.abc import Collection, Iterable, Mapping, Sequence
from typing import Any, Generic, TypeVar

import numpy as np
from aviary.message import Message
from aviary.tools import Tool, ToolRequestMessage
from pydantic import BaseModel, ConfigDict, Field, JsonValue

from ldp.graph import Op, OpResult

try:
    # So we can skip torch objects when looking for Ops
    import torch
except ImportError:
    # If torch is not available, then it won't be used in an Agent anyway
    torch = None  # type: ignore[assignment]

TAgentState = TypeVar("TAgentState")


# A global registry of all Agent subclasses, so we can look them up by name
_AGENT_REGISTRY: dict[str, type[Agent]] = {}


class Agent(ABC, Generic[TAgentState]):
    def __init_subclass__(cls, **kwargs):
        """Ensure Ops have unique names and subclasses are in _AGENT_REGISTRY."""
        super().__init_subclass__(**kwargs)

        original_init = cls.__init__

        def init_with_op_naming(self, *args, **kwargs):
            original_init(self, *args, **kwargs)

            # loop through Ops and give them proper names
            for name, op in _find_ops(self):
                op.set_name(name)

        cls.__init__ = init_with_op_naming  # type: ignore[method-assign]

        # Register the Agent subclass.
        _AGENT_REGISTRY[cls.__name__] = cls

    @abstractmethod
    async def get_asv(
        self, agent_state: TAgentState, obs: list[Message]
    ) -> tuple[OpResult[ToolRequestMessage], TAgentState, float]:
        """
        Get new action, state, and value given state and observation messages.

        NOTE: the method's name has action listed before state to help you realize it's
        a new state.

        Args:
            agent_state: Optional current agent state, pass None if irrelevant.
                This can be something like agent memory.
            obs: Most recent list of observation messages from the environment's steps.
                If more observations than the most recent list are necessary, track them
                in the agent state.

        Returns:
            Three-tuple of new action, new agent state, and estimated value. The
                agent_state is returned as a copy so that you can safely mutate it
                without affecting the original. The estimated value is the agent's
                estimate of the future rewards given the input state and observations,
                and is used for RL training.  If estimated value doesn't matter, just
                return 0. The value could also come from a Q-value evaluated at the
                action chosen by the agent.
        """

    @abstractmethod
    async def init_state(self, tools: list[Tool]) -> TAgentState:
        """Initializes the first agent state with the provided tools."""

    def named_ops(self) -> Iterable[tuple[str, Op]]:
        """Analogous to torch.nn.Module.named_parameters()."""
        return _find_ops(self)

    @classmethod
    def from_name(cls, name: str, **kwargs) -> Agent:
        return _AGENT_REGISTRY[name](**kwargs)


class AgentConfig(BaseModel):
    """Configuration for specifying the type of agent i.e. the subclass of Agent above."""

    model_config = ConfigDict(extra="forbid")

    agent_type: str = Field(
        description="The type of agent to be used. "
        "This should be a subclass of Agent above.",
    )
    agent_kwargs: dict[str, JsonValue] = Field(
        default_factory=dict,
        description="Keyword arguments to pass to the agent's constructor.",
    )

    def construct_agent(self) -> Agent:
        return Agent.from_name(self.agent_type, **self.agent_kwargs)

    def __hash__(self) -> int:
        return hash(self.agent_type + json.dumps(self.agent_kwargs, sort_keys=True))


def _find_ops(  # noqa: C901
    root: object, root_name: str = "", visited: set[int] | None = None
) -> Iterable[tuple[str, Op]]:
    """Recursive function to find children that are Ops and the attr chain to reach them.

    E.g. if root.module.op is an Op, then we will yield ("module.op", root.module.op).
    These are not fully qualified names, but more like "locally qualified names". In the above
    example, "root." + "module.op" is the fully qualified name.
    This is an internal function - Agent.named_ops() should usually suffice.

    Args:
        root: Any object that might hold Ops.
        root_name: The name of the root object. Defaults to empty string and is passed as an arg to
            make this method recursive.
        visited: a set of visited node IDs to avoid loops. Defaults to None.

    Yields:
        Two-tuple of (locally qualified name, Op) pairs
    """
    # Recursive function to find children that are Ops and the
    # attribute chain to reach them.
    if visited is None:
        visited = set()

    if isinstance(root, Op):
        yield root_name, root
        # Assume an Op may not have sub-Ops. I think this is sound, since
        # we wouldn't be tracking the compute graph properly if it did.
        return

    if "__pydantic_parent_namespace__" in root_name:
        # Skip Pydantic internals
        return

    # Don't recurse into PyTorch objects because they won't contain Ops
    if torch is not None and (  # type: ignore[redundant-expr]
        isinstance(root, torch.Tensor | torch.nn.Module)
    ):
        return

    # Similarly for numpy
    if isinstance(root, np.ndarray):
        return

    # loop through 3 types of containers: dicts, collections, and objects
    if isinstance(root, Mapping):
        named_attrs: Any = root.items()
    elif isinstance(root, Sequence | Collection) and not isinstance(root, str | bytes):
        named_attrs = enumerate(root)
    elif hasattr(root, "__dict__"):
        # object?
        named_attrs = root.__dict__.items()
    else:
        # couldn't descend
        return

    for k, v in named_attrs:
        id_v = id(v)
        if id_v not in visited:
            # only visit each object once - avoid loops, etc.
            visited.add(id_v)
            if root_name:
                k = f"{root_name}.{k}"
            yield from _find_ops(v, root_name=k, visited=visited)



import asyncio
import uuid
from collections.abc import Awaitable, Callable, Sequence
from contextlib import suppress
from typing import NamedTuple

from aviary.env import Environment

from ldp.agent.agent import Agent, TAgentState
from ldp.data_structures import Trajectory, Transition

from .callbacks import Callback
from .rollout import AgentError, EnvError, TEnv, reraise_exc_as


class Beam(NamedTuple):
    # An ongoing beam contains two things: the trajectory up to now
    # and the environment that the last action was sampled from. We
    # need both to continue sampling the next step.
    traj: Trajectory
    env: Environment


class BeamSearchRollout:
    def __init__(
        self,
        agent: Agent,
        beam_width: int,
        samples_per_beam: int,
        env_clone_fn: Callable[[TEnv], Awaitable[TEnv]],
        agent_clone_fn: Callable[[TAgentState], TAgentState],
        scoring_fn: Callable[[Trajectory], Awaitable[float]],
        replay_actions_on_clone: bool = False,
        callbacks: Sequence[Callback] | None = None,
        catch_agent_failures: bool = True,
        catch_env_failures: bool = True,
        verbose: bool = False,
    ):
        self.agent = agent

        self.catch_agent_failures = catch_agent_failures
        self.catch_env_failures = catch_env_failures

        self.verbose = verbose

        self.traj_buffer: dict[str, Trajectory] = {}
        self.search_buffer: dict[str, list[Trajectory]] = {}

        self.beam_width = beam_width
        self.samples_per_beam = samples_per_beam

        self.env_clone_fn = env_clone_fn
        self.agent_clone_fn = agent_clone_fn
        self.scoring_fn = scoring_fn
        self.replay_actions_on_clone = replay_actions_on_clone

        self.callbacks = callbacks or []

    async def sample_trajectories(
        self,
        environments: Sequence[Environment],
        max_steps: int | None = None,
    ) -> list[Trajectory]:
        self.traj_buffer.clear()
        traj_ids = [uuid.uuid4().hex for _ in environments]

        tasks = [
            self._rollout(traj_id, env, max_steps)
            for traj_id, env in zip(traj_ids, environments, strict=True)
        ]
        await asyncio.gather(*tasks)

        return [self.traj_buffer[traj_id] for traj_id in traj_ids]

    async def _rollout(
        self, traj_id: str, env: Environment, max_steps: int | None
    ) -> None:
        with suppress(AgentError, EnvError):
            # for samples_per_beam==1. we want to ensemble and pick the highest-scoring one
            n_seeds = 1 if self.samples_per_beam > 1 else self.beam_width

            done_beams: list[Beam] = []
            beams = [
                Beam(traj=Trajectory(traj_id=f"{traj_id}:{i}"), env=env)
                for i in range(n_seeds)
            ]
            # will be replaced if rollout is successful
            self.traj_buffer[traj_id] = beams[0].traj
            self.search_buffer[traj_id] = []

            with reraise_exc_as(EnvError, self.catch_env_failures):
                init_obs, tools = await env.reset()
            await asyncio.gather(*[
                c.after_env_reset(traj_id, init_obs, tools) for c in self.callbacks
            ])

            with reraise_exc_as(AgentError, self.catch_agent_failures):
                seed_agent_states = await asyncio.gather(
                    *(self.agent.init_state(tools) for _ in range(n_seeds))
                )
            # TODO: implement after_agent_init_state callback

            while len(done_beams) < self.beam_width and beams:
                new_beams = []
                for beam, seed_agent_state in zip(
                    beams, seed_agent_states, strict=True
                ):
                    for i_sample in range(self.samples_per_beam):
                        new_env = await self._clone_env(beam)
                        if new_env is None:
                            continue

                        agent_state = self.agent_clone_fn(seed_agent_state)
                        obs = (
                            beam.traj.steps[-1].next_observation
                            if beam.traj.steps
                            else init_obs.copy()
                        )

                        await asyncio.gather(*[
                            callback.before_transition(
                                traj_id, self.agent, env, agent_state, obs
                            )
                            for callback in self.callbacks
                        ])

                        with reraise_exc_as(AgentError, self.catch_agent_failures):
                            (
                                action,
                                next_agent_state,
                                vhat,
                            ) = await self.agent.get_asv(agent_state, obs)
                        await asyncio.gather(*[
                            callback.after_agent_get_asv(
                                traj_id, action, next_agent_state, vhat
                            )
                            for callback in self.callbacks
                        ])

                        with reraise_exc_as(EnvError, self.catch_env_failures):
                            next_obs, reward, done, trunc = await new_env.step(
                                action.value
                            )
                        await asyncio.gather(*[
                            callback.after_env_step(
                                traj_id, next_obs, reward, done, trunc
                            )
                            for callback in self.callbacks
                        ])

                        step = Transition(
                            timestep=len(beam.traj.steps),
                            agent_state=agent_state,
                            next_agent_state=next_agent_state,
                            observation=obs,
                            next_observation=next_obs,
                            action=action,
                            reward=reward,
                            done=done,
                            truncated=trunc,
                            value=0.0,  # will be filled in
                        )
                        await asyncio.gather(*[
                            callback.after_transition(traj_id, self.agent, env, step)
                            for callback in self.callbacks
                        ])

                        new_beam = Beam(
                            traj=Trajectory(
                                traj_id=beam.traj.traj_id + f":{i_sample}",  # type: ignore[operator]
                                steps=[*beam.traj.steps, step],
                            ),
                            env=new_env,
                        )
                        step.value = await self.scoring_fn(new_beam.traj)
                        self.search_buffer[traj_id].append(new_beam.traj)

                        if (
                            not new_beam.traj.done
                            and max_steps is not None
                            and len(new_beam.traj.steps) >= max_steps
                        ):
                            last_step = new_beam.traj.steps[-1]
                            last_step.done = last_step.truncated = True

                        if new_beam.traj.done:
                            done_beams.append(new_beam)
                        else:
                            new_beams.append(new_beam)

                new_beams.sort(key=lambda b: b.traj.steps[-1].value, reverse=True)
                beams, discarded = (
                    new_beams[: self.beam_width],
                    new_beams[self.beam_width :],
                )
                seed_agent_states = [b.traj.steps[-1].next_agent_state for b in beams]
                await asyncio.gather(*[d.env.close() for d in discarded])

            await asyncio.gather(*[b.env.close() for b in beams])

            self.traj_buffer[traj_id] = max(
                done_beams,
                key=lambda b: (b.traj.steps[-1].truncated, b.traj.steps[-1].value),
            ).traj

    async def _clone_env(self, beam: Beam) -> Environment | None:
        try:
            with reraise_exc_as(EnvError, self.catch_env_failures):
                # I'm not sure how to type hint this
                env = await self.env_clone_fn(beam.env)  # type: ignore[arg-type]
                if self.replay_actions_on_clone:
                    # Some envs can't be cloned, so instead replay.
                    # We rely on env_clone_fn to properly reset the env if needed.
                    # We assume a deterministic env, so the return values are discarded.
                    for step in beam.traj.steps:
                        if step.action is not None:
                            _ = await env.step(step.action.value)
                return env
        except EnvError:
            return None



import json
import logging
import os
import time
from collections import defaultdict
from collections.abc import Collection, Iterable, Sequence
from pathlib import Path
from typing import Any, cast

import aiofiles
from aviary.env import Environment, TaskDataset
from aviary.message import Message
from aviary.tools import MessagesAdapter, Tool, ToolRequestMessage

from ldp.agent import Agent
from ldp.data_structures import Trajectory, Transition
from ldp.graph import OpCtx, OpResult

try:
    import wandb
except ImportError:
    wandb = None  # type: ignore[assignment]

logger = logging.getLogger(__name__)


class Callback:
    """Base class for callbacks used by RolloutManager/Evaluator/OnlineTrainer.

    Pseudocode to demonstrate how callback methods are invoked (marked as *):

    RolloutManager.sample_trajectories():
        env.reset()
        callback.after_env_reset() *
        agent.init_state()
        callback.after_agent_init_state() *
        while not done:
            callback.before_transition() *
            agent.get_asv()
            callback.after_agent_get_asv() *
            env.step()
            callback.after_env_step() *
            callback.after_transition() *

    Evaluator.evaluate / OnlineTrainer._eval_loop():
        callback.before_eval_loop() *
        for batch in eval_dataset:
            rollout_manager.sample_trajectories()
            callback.after_eval_step() *
        callback.after_eval_loop() *

    OfflineTrainer / OnlineTrainer.train():
        for batch in train_dataset:
            rollout_manager.sample_trajectories() # if online
            optimizer.aggregate()
            if updating_optimizer:
                optimizer.update()
                callback.after_update() *
            callback.after_train_step() *
    """

    async def before_transition(
        self,
        traj_id: str,
        agent: Agent,
        env: Environment,
        agent_state: Any,
        obs: list[Message],
    ) -> None:
        """Invoked by runners before each transition and after agent and env reset."""

    async def after_agent_init_state(self, traj_id: str, init_state: Any) -> None:
        """Invoked by runners after agent.init_state()."""

    async def after_agent_get_asv(
        self,
        traj_id: str,
        action: OpResult[ToolRequestMessage],
        next_agent_state: Any,
        value: float,
    ) -> None:
        """Invoked by runners after agent.get_asv()."""

    async def after_env_reset(
        self, traj_id: str, obs: list[Message], tools: list[Tool]
    ) -> None:
        """Invoked by runners after env.reset()."""

    async def after_env_step(
        self, traj_id: str, obs: list[Message], reward: float, done: bool, trunc: bool
    ) -> None:
        """Invoked by runners after env.step()."""

    async def after_transition(
        self, traj_id: str, agent: Agent, env: Environment, transition: Transition
    ) -> None:
        """Invoked by runners after each transition."""

    async def after_train_step(self, trajectories: Sequence[Trajectory]) -> None:
        """Invoked by OnlineTrainer after each training step."""

    async def before_eval_loop(self) -> None:
        """Invoked by Evaluator and OnlineTrainer before the evaluation loop."""

    async def after_eval_step(self, trajectories: Sequence[Trajectory]) -> None:
        """Invoked by Evaluator and OnlineTrainer after each evaluation step."""

    async def after_eval_loop(self) -> None:
        """Invoked by Evaluator and OnlineTrainer after the evaluation loop."""

    async def after_update(self) -> None:
        """Invoked by OnlineTrainer after each optimizer.update() call."""


class TrajectoryFileCallback(Callback):
    """Callback that writes trajectories to a file."""

    def __init__(self, output_dir: os.PathLike | str):
        self.output_dir = Path(output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)

        self.out_files: dict[str, Path] = {}
        self.trajs: dict[str, Trajectory] = defaultdict(Trajectory)

    def _make_filename(self, traj_id: str, env: Environment) -> str:
        """Create the filename for the output file."""
        return f"{traj_id}.jsonl"

    async def before_transition(
        self,
        traj_id: str,
        agent: Agent,
        env: Environment,
        agent_state: Any,
        obs: list[Message],
    ) -> None:
        if traj_id not in self.out_files:
            self.out_files[traj_id] = self.output_dir / self._make_filename(
                traj_id, env
            )

    async def after_transition(
        self, traj_id: str, agent: Agent, env: Environment, transition: Transition
    ) -> None:
        assert traj_id in self.out_files
        traj = self.trajs[traj_id]
        traj.steps.append(transition)
        # TODO: make this async?
        traj.to_jsonl(self.out_files[traj_id])

    def cleanup(self) -> None:
        for out_file in self.out_files.values():
            if out_file.exists():
                out_file.unlink()


class RolloutDebugDumpCallback(Callback):
    """Dump JSONL files for each agent and environment step to a directory."""

    def __init__(self, output_dir: os.PathLike | str):
        """Initialize.

        Args:
            output_dir: Directory to place JSONL files.
        """
        self.output_dir = Path(output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)

        self.out_files: dict[str, Path] = {}

    def _get_out_file(self, traj_id: str) -> Path:
        if traj_id not in self.out_files:
            self.out_files[traj_id] = self.output_dir / f"{traj_id}.jsonl"
        return self.out_files[traj_id]

    async def before_transition(
        self,
        traj_id: str,
        agent: Agent,
        env: Environment,
        agent_state,
        obs: list[Message],
    ) -> None:
        self.start = time.time()

    def _get_elapsed_time(self, reset: bool = True) -> float:
        elapsed = time.time() - self.start
        if reset:
            self.start = time.time()
        return elapsed

    async def after_agent_get_asv(
        self,
        traj_id: str,
        action: OpResult[ToolRequestMessage],
        next_agent_state: Any,
        value: float,
    ) -> None:
        log_jsonl = json.dumps({
            "event": "AGENT_GET_ASV",
            "elapsed": self._get_elapsed_time(),
            "action": action.value.model_dump(),
            "value": value,
        })
        async with aiofiles.open(self._get_out_file(traj_id), "a") as f:
            await f.write(log_jsonl + "\n")

    async def after_env_step(
        self, traj_id: str, obs: list[Message], reward: float, done: bool, trunc: bool
    ) -> None:
        log_jsonl = json.dumps({
            "event": "ENV_STEP",
            "elapsed": self._get_elapsed_time(),
            "obs": MessagesAdapter.dump_python(obs),
            "reward": reward,
            "done": done,
            "truncated": trunc,
        })
        async with aiofiles.open(self._get_out_file(traj_id), "a") as f:
            await f.write(log_jsonl + "\n")


class ComputeTrajectoryMetricsMixin:
    """Mixin for TaskDataset classes to enable them to compute metrics."""

    # Tools or tool names to include in trajectory metrics
    tools_to_track: Collection[str | Tool] = set()

    def compute_trajectory_metrics(
        self,
        trajectories: Sequence[Trajectory],
    ) -> dict[str, list[float]]:
        metrics: dict[str, list[float]] = {
            "reward": [
                sum(step.reward for step in traj.steps) for traj in trajectories
            ],
            "truncation_rate": [
                sum(step.truncated for step in traj.steps) for traj in trajectories
            ],
            "avg_value": [
                sum(step.value for step in traj.steps) / len(traj.steps)
                for traj in trajectories
            ],
            "num_steps": [len(traj.steps) for traj in trajectories],
            "failures": [traj.failed for traj in trajectories],
        }
        for tool in self.tools_to_track:  # Default of empty set means this is not run
            if isinstance(tool, Tool):
                tool = tool.info.name
            metrics[f"tool_{tool}"] = [
                sum(
                    sum(tc.function.name == tool for tc in s.action.value.tool_calls)
                    for s in traj.steps
                    if isinstance(s.action, OpResult)
                )
                for traj in trajectories
            ]
        return metrics


class TrajectoryMetricsCallback(Callback):
    """
    Compute metrics that are defined by task datasets.

    NOTE: evaluation portion's after_eval_step/loop() is not concurrency safe because
    trajectories should be stored in the order of after_eval_step() calls.
    """

    def __init__(
        self,
        train_dataset: TaskDataset | None = None,
        eval_dataset: TaskDataset | None = None,
        track_tool_usage: bool = False,
    ):
        self._datasets = train_dataset, eval_dataset
        self._track_tool_usage = track_tool_usage
        for ds in self._datasets:
            if ds and not isinstance(ds, ComputeTrajectoryMetricsMixin):
                raise ValueError(
                    f"Dataset {ds} didn't implement"
                    f" {ComputeTrajectoryMetricsMixin.__name__}, which is required for"
                    " this callback."
                )
        self._train_metrics_fn = (
            train_dataset.compute_trajectory_metrics if train_dataset else None  # type: ignore[attr-defined]
        )
        self._eval_metrics_fn = (
            eval_dataset.compute_trajectory_metrics if eval_dataset else None  # type: ignore[attr-defined]
        )

        self._train_metrics: dict[str, list[float]] | None = None
        self._eval_metrics: dict[str, list[float]] = {}

    async def after_env_reset(
        self, traj_id: str, obs: list[Message], tools: list[Tool]
    ) -> None:
        for ds in (ds for ds in self._datasets if ds):
            if self._track_tool_usage:
                cast(ComputeTrajectoryMetricsMixin, ds).tools_to_track = {
                    t.info.name for t in tools
                }

    async def after_train_step(self, trajectories: Sequence[Trajectory]) -> None:
        if self._train_metrics_fn is not None:
            self._train_metrics = self._train_metrics_fn(trajectories)

    async def after_eval_step(self, trajectories: Sequence[Trajectory]) -> None:
        if self._eval_metrics_fn is not None:
            for k, v in self._eval_metrics_fn(trajectories).items():
                if k not in self._eval_metrics:
                    # Don't use defaultdict - error prone in user code
                    self._eval_metrics[k] = []
                self._eval_metrics[k].extend(v)

    async def after_eval_loop(self) -> None:
        self._eval_metrics.clear()


class MeanMetricsCallback(TrajectoryMetricsCallback):
    """Take a mean of all metrics."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._train_means: dict[str, float] | None = None
        self._eval_means: dict[str, float] | None = None

    async def after_train_step(self, trajectories: Sequence[Trajectory]) -> None:
        await super().after_train_step(trajectories)
        if self._train_metrics is not None:
            # may be None if train_dataset was not provided
            self._train_means = self._compute_means(self._train_metrics)

    async def after_eval_loop(self) -> None:
        if self._eval_metrics:
            # may be empty if eval_dataset was not provided
            self._eval_means = self._compute_means(self._eval_metrics)
        await super().after_eval_loop()

    @staticmethod
    def _compute_means(metrics: dict[str, list[float]]) -> dict[str, float]:
        return {k: sum(v) / len(v) for k, v in metrics.items()}

    @property
    def train_means(self) -> dict[str, float]:
        if self._train_means is None:
            raise RuntimeError(
                "Training means are only available after this callback is invoked."
            )
        return self._train_means

    @property
    def eval_means(self) -> dict[str, float]:
        if self._eval_means is None:
            raise RuntimeError(
                "Evaluation means are only available after this callback is invoked."
            )
        return self._eval_means


class WandBLoggingCallback(TrajectoryMetricsCallback):
    def __init__(self, *args, **kwargs):
        if wandb is None:
            raise ImportError(
                f"{type(self).__name__} processing requires the 'monitor' extra for"
                " 'wandb'. Please: `pip install aviary-internal[monitor]`."
            )
        super().__init__(*args, **kwargs)

        self._num_train_step = 0

    async def after_train_step(self, trajectories: Sequence[Trajectory]) -> None:
        await super().after_train_step(trajectories)
        self._num_train_step += 1

        if self._train_metrics is None:
            return

        # Each wandb.log() increments the wandb step by 1. Log the training step here
        # so we can use it as an x-axis for training metrics that are logged by different
        # wandb.log() calls.
        wandb.log(
            {
                f"train/{key}_mean": sum(vals) / len(vals)
                for key, vals in self._train_metrics.items()
            }
            | {"train/step": self._num_train_step}
        )

    async def after_eval_loop(self) -> None:
        if not self._eval_metrics:
            return

        wandb.log({
            f"eval/{key}_mean": (sum(vals) / len(vals) if vals else None)
            for key, vals in self._eval_metrics.items()
        })

        await super().after_eval_loop()


class ClearContextCallback(Callback):
    def __init__(self, op_names: Iterable[str] | None = None):
        self._op_names = op_names

    async def after_eval_step(self, trajectories: Sequence[Trajectory]) -> None:
        OpCtx.clear_contexts(self._op_names)

    async def after_update(self) -> None:
        OpCtx.clear_contexts(self._op_names)


class LoggingCallback(MeanMetricsCallback):
    """Custom callback for logging filtered metrics (e.g., pass rates) to the console.

    This callback extends the `MeanMetricsCallback` and allows logging of user-specified metrics
    after each training step and after the evaluation loop. It calculates the specified metrics
    (e.g., pass rates) from the trajectories and logs the results.
    """

    def __init__(
        self,
        train_dataset: TaskDataset | None = None,
        eval_dataset: TaskDataset | None = None,
        metrics_to_log: Collection[str] | None = None,
    ):
        """Initialize the callback with a list of metric keys to log.

        Args:
            train_dataset: The training dataset for computing metrics.
            eval_dataset: The evaluation dataset for computing metrics.
            metrics_to_log: Optional metric names (e.g., ["pass"]) to log.
                            If left as default of None, all metrics will be logged.
        """
        super().__init__(train_dataset, eval_dataset)
        self.metrics_to_log = (
            metrics_to_log or set()
        )  # If no metrics provided, log all by default

    def _log_filtered_metrics(self, metrics: dict[str, float], step_type: str) -> None:
        """Helper function to log only the specified metrics.

        Args:
            metrics: Dictionary of calculated means for the current step (e.g., train or eval).
            step_type: The type of step (e.g., "Train" or "Eval") for logging purposes.
        """
        if self.metrics_to_log:
            for metric in self.metrics_to_log:
                if metric in metrics:
                    logger.info(
                        f"{metric.upper()} RATE ({step_type}): {metrics[metric]:.5f}"
                    )
        else:
            # Log all metrics if no specific ones are provided
            logger.info(f"{step_type} Metrics: {metrics}")

    async def after_train_step(self, trajectories: Sequence[Trajectory]) -> None:
        """Log metrics and pass rate after each training step.

        This method is called after every training step, calculating and logging
        the training metrics and pass rate.

        Args:
            trajectories: A sequence of trajectories from the training step.
        """
        await super().after_train_step(trajectories)  # Call the parent to compute means
        if self.train_means:
            self._log_filtered_metrics(self.train_means, step_type="Train")

    async def after_eval_loop(self) -> None:
        """Log metrics and pass rate after the evaluation loop.

        This method is called after the evaluation loop finishes, calculating and logging
        the evaluation metrics and pass rate.
        """
        await super().after_eval_loop()  # Call the parent to compute means
        if self.eval_means:
            self._log_filtered_metrics(self.eval_means, step_type="Eval")


class TerminalPrintingCallback(Callback):
    """Callback that prints action, observation, and timing information to the terminal."""

    def __init__(self):
        self.start_time = None
        # try now, rather than start running and die
        try:
            from rich.pretty import pprint  # noqa: F401
        except ImportError as e:
            raise ImportError(
                f"rich is required for {type(self).__name__}. Please install it with `pip install rich`."
            ) from e

    async def before_transition(
        self,
        traj_id: str,
        agent: Agent,
        env: Environment,
        agent_state: Any,
        obs: list[Message],
    ) -> None:
        """Start the timer before each transition."""
        self.start_time = time.time()

    async def after_agent_get_asv(
        self,
        traj_id: str,
        action: OpResult[ToolRequestMessage],
        next_agent_state: Any,
        value: float,
    ) -> None:
        from rich.pretty import pprint

        print("\nAction:")
        pprint(action.value, expand_all=True)

    async def after_env_step(
        self, traj_id: str, obs: list[Message], reward: float, done: bool, trunc: bool
    ) -> None:
        from rich.pretty import pprint

        # Compute elapsed time
        if self.start_time is not None:
            elapsed_time = time.time() - self.start_time
            self.start_time = None  # Reset timer
        else:
            elapsed_time = 0.0
        print("\nObservation:")
        pprint(obs, expand_all=True)
        print(f"Elapsed time: {elapsed_time:.2f} seconds")



import asyncio
from abc import ABC, abstractmethod
from enum import StrEnum
from typing import Any

import litellm
import numpy as np
import tiktoken
from pydantic import BaseModel, ConfigDict, Field, model_validator


class EmbeddingModes(StrEnum):
    """Enum representing the different modes of an embedding model."""

    DOCUMENT = "document"
    QUERY = "query"


class EmbeddingModel(ABC, BaseModel):
    name: str
    dimensions: int | None = None

    def set_mode(self, mode: EmbeddingModes) -> None:
        """Several embedding models have a 'mode' or prompt which affects output."""

    @abstractmethod
    async def embed_texts(self, texts: list[str]) -> list[np.ndarray]:
        pass

    async def embed_text(self, text: str) -> np.ndarray:
        return (await self.embed_texts([text]))[0]

    @staticmethod
    def from_name(embedding: str, **kwargs) -> "EmbeddingModel":
        if embedding.startswith("hybrid"):
            dense_model = LiteEmbeddingModel(name="-".join(embedding.split("-")[1:]))
            return HybridEmbeddingModel(
                name=embedding, models=[dense_model, SparseEmbeddingModel(**kwargs)]
            )
        if embedding == "sparse":
            return SparseEmbeddingModel(**kwargs)
        return LiteEmbeddingModel(name=embedding, **kwargs)


class LiteEmbeddingModel(EmbeddingModel):
    name: str = Field(default="text-embedding-3-small")
    dimensions: int | None = Field(
        default=None,
        description=(
            "The length an embedding will have. If left unspecified, we attempt to"
            " infer an un-truncated length via LiteLLM's internal model map. If this"
            " inference fails, the embedding will be un-truncated."
        ),
    )
    batch_size: int = 16
    embed_kwargs: dict[str, Any] = Field(
        default_factory=dict,
        description="Extra kwargs to pass to litellm.aembedding.",
    )

    @model_validator(mode="before")
    @classmethod
    def infer_dimensions(cls, data: dict[str, Any]) -> dict[str, Any]:
        if data.get("dimensions") is not None:
            return data
        # Let's infer the dimensions
        config: dict[str, dict[str, Any]] = litellm.get_model_cost_map(
            url="https://raw.githubusercontent.com/BerriAI/litellm/main/litellm/model_prices_and_context_window_backup.json"
        )
        output_vector_size: int | None = config.get(data.get("name", ""), {}).get(  # noqa: FURB184
            "output_vector_size"
        )
        if output_vector_size:
            data["dimensions"] = output_vector_size
        return data

    async def embed_texts(self, texts: list[str]) -> list[np.ndarray]:
        embeddings = []
        # Before you get excited to asyncio.gather this:
        # The point of this is to not hit the API rate limit
        for i in range(0, len(texts), self.batch_size):
            response = await litellm.aembedding(
                model=self.name,
                input=texts[i : i + self.batch_size],
                encoding_format="float",
                dimensions=self.dimensions,
                **self.embed_kwargs,
            )
            embeddings.extend([
                np.array(e["embedding"], dtype=np.float32) for e in response.data
            ])
        return embeddings


class SparseEmbeddingModel(EmbeddingModel):
    """This is a very simple keyword search model - probably best to be mixed with others."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str = "sparse"
    dimensions: int = 256
    enc: tiktoken.Encoding = Field(
        default_factory=lambda: tiktoken.get_encoding("cl100k_base")
    )

    async def embed_texts(self, texts) -> list[np.ndarray]:
        enc_batch = self.enc.encode_ordinary_batch(texts)
        # now get frequency of each token rel to length
        return [
            np.bincount(
                [xi % self.dimensions for xi in x], minlength=self.dimensions
            ).astype(np.float32)
            / len(x)
            for x in enc_batch
        ]


class HybridEmbeddingModel(EmbeddingModel):
    name: str = "hybrid-embed"
    models: list[EmbeddingModel]

    @model_validator(mode="before")
    @classmethod
    def infer_dimensions(cls, data: dict[str, Any]) -> dict[str, Any]:
        if data.get("dimensions") is not None:
            raise ValueError(f"Don't specify dimensions to {cls.__name__}.")
        if not data.get("models") or any(m.dimensions is None for m in data["models"]):
            return data
        data["dimensions"] = sum(m.dimensions for m in data["models"])
        return data

    async def embed_texts(self, texts):
        all_embeds = await asyncio.gather(*[m.embed_texts(texts) for m in self.models])
        return np.concatenate(all_embeds, axis=1)



__all__ = ["AsyncTorchModule", "async_protect_torch_call"]

import asyncio
import operator
import time
from abc import ABC, abstractmethod
from collections.abc import Callable
from contextlib import nullcontext
from typing import Any
from uuid import UUID, uuid4

try:
    import torch
    from torch import nn
    from torch.utils.data import default_collate
except ImportError:
    raise ImportError(
        "ldp.graph.async_torch requires PyTorch as a dependency. "
        "Please run `pip install ldp[nn]`."
    ) from None

_TORCH_LOCK = asyncio.Lock()

# Supported devices here: https://pytorch.org/docs/stable/amp.html#torch.autocast
_AUTOCAST_DEVICES = {"cpu", "cuda", "hpu", "xpu"}


def _get_autocast_context(dtype: torch.dtype | None, device_type: str):
    return (
        nullcontext()
        if dtype is None or device_type not in _AUTOCAST_DEVICES
        else torch.autocast(dtype=dtype, device_type=device_type)
    )


def _get_grad_context(no_grad: bool):
    return torch.no_grad() if no_grad else nullcontext()


def async_protect_torch_call(
    module: nn.Module,
    module_call_fn: Callable = lambda m, *args, **kwargs: m(*args, **kwargs),
    no_grad: bool = False,
    autocast_dtype: torch.dtype | None = None,
    autocast_device_type=None,
) -> Callable:
    async def wrapped_call(*args, **kwargs):
        async with _TORCH_LOCK:
            with (
                _get_grad_context(no_grad),
                _get_autocast_context(autocast_dtype, autocast_device_type),
            ):
                return module_call_fn(module, *args, **kwargs)

    return wrapped_call


# TODO: make max_wait_interval adaptive. We can use a heuristic like
# half the average time for a single call. If it's not provided, enable
# adaptive mode.


class AsyncBufferedWorker(ABC):
    """Abstract class for a worker that buffers inputs and processes them in batches."""

    def __init__(
        self,
        batch_size: int,
        max_wait_interval: float,
        collate_fn: Callable = lambda x: x,
        decollate_fn: Callable = lambda x: x,
    ):
        """Initialize.

        Args:
            batch_size: The target batch size to use when calling the module. As soon as
                batch_size calls are made, a forward pass is executed.
            max_wait_interval: The maximum time (sec) to wait for a batch to fill up
                before executing the calls we have buffered.
            collate_fn: A function to pre-process a list of inputs into a batch. Defaults to a
                no-op.
            decollate_fn: Kind of like the opposite of collate_fn. This function should take
                 the batched output and return an ordered list of outputs. Defaults to no-op.
        """
        self.batch_size = batch_size
        self.timeout = max_wait_interval
        self.collate_fn = collate_fn
        self.decollate_fn = decollate_fn

        self._work_buffer: list[tuple[float, UUID, dict[str, Any]]] = []
        self._result_buffer: dict[UUID, Any] = {}
        self._lock = asyncio.Lock()

    async def __call__(self, **kwargs):
        request_id = uuid4()
        request_ts = time.time()

        async with self._lock:
            # Make sure only one coroutine is using the work buffer at a time
            self._work_buffer.append((request_ts, request_id, kwargs))

        while True:
            async with self._lock:
                # Only one coroutine allowed in here when:
                # - modifying the result buffer
                # - modifying the work buffer

                if request_id in self._result_buffer:
                    # Our request was fulfilled by this or another coroutine!
                    return self._result_buffer.pop(request_id)

                # Try to run a batch.
                await self._maybe_process_batch()

            # Sleep, to let another coroutine take over if it needs to
            await asyncio.sleep(0.0)

    async def _maybe_process_batch(self):
        """If the buffer is >= batch size or we have been waiting long enough, process the old batch.

        If neither condition is met, do nothing.
        """
        now = time.time()

        # sort by oldest requests first
        self._work_buffer.sort(key=operator.itemgetter(0))

        if (
            len(self._work_buffer) >= self.batch_size
            or now - self._work_buffer[0][0] > self.timeout
        ):
            # if we're over batch size or have at least one input waiting for
            # more than timeout, pull out a batch to run
            batch = self._work_buffer[: self.batch_size]
            self._work_buffer = self._work_buffer[self.batch_size :]

            # Construct the batch tensors
            sample_kwargs = [x[2] for x in batch]
            batch_kwargs = self.collate_fn(sample_kwargs)

            batched_results = await self._batched_call(batch_kwargs)
            request_ids = [x[1] for x in batch]
            results = self.decollate_fn(batched_results)
            self._result_buffer.update(zip(request_ids, results, strict=True))

    @abstractmethod
    async def _batched_call(self, batch_kwargs: dict[str, Any]):
        """Logic to call the worker on a batch of inputs."""


class AsyncTorchModule(AsyncBufferedWorker):
    def __init__(
        self,
        module: nn.Module,
        batch_size: int,
        max_wait_interval: float,
        collate_fn: Callable = default_collate,
        decollate_fn: Callable = list,
        module_call_fn: Callable = lambda m, *args, **kwargs: m(*args, **kwargs),
    ):
        """A wrapper around a torch.nn.Module that allows for async calls.

        Usage:
        ```python
        my_model = nn.Linear(2, 2)
        async_model = AsyncTorchModule(my_model, batch_size=4, max_wait_interval=0.01)

        result = await asyncio.gather(*[
            async_model(input=torch.rand(2)) for _ in range(10)
        ])
        ```
        In the above example, note that we are making 10 calls with a batch size of 4.
        The first two groups of 4 will be batched and executed as they arrive. The last 2
        will wait for max_wait_interval and then execute.

        NOTE: This module is not thread-safe and currently always operates in no_grad() mode.
        It may be possible to relax the latter constraint.

        Args:
            module: The PyTorch module to wrap.
            batch_size: See parent class.
            max_wait_interval: See parent class.
            collate_fn: A PyTorch collate function to use when batching inputs. Defaults to
                the PyTorch default_collate.
            decollate_fn: Kind of like the opposite of collate_fn. This function should take
                 the batched output and return an ordered list of outputs. Defaults to list.
            module_call_fn: Function that allows for customizing the call to the module.
        """
        super().__init__(
            batch_size=batch_size,
            max_wait_interval=max_wait_interval,
            collate_fn=collate_fn,
            decollate_fn=decollate_fn,
        )
        self.module = module
        self.module_call_fn = module_call_fn

    def _get_dtype_and_device(self) -> tuple[torch.dtype, torch.device]:
        param = next(self.module.parameters())
        return param.dtype, param.device

    async def _batched_call(self, batch_kwargs: dict[str, Any]):
        # Call the module and store results
        # To be safe, set _TORCH_LOCK to prevent other
        # coroutines from messing with torch state while running.
        async with _TORCH_LOCK:
            dtype, device = self._get_dtype_and_device()
            with torch.no_grad(), _get_autocast_context(dtype, device.type):
                return self.module_call_fn(self.module, **batch_kwargs)



"""This module provides utility functions for appending and prepending system messages."""

from collections.abc import Collection, Iterable

from aviary.message import Message


def append_to_messages(messages: list[Message], new_message: Message) -> list[Message]:
    """Appends a message to a list of messages, returning that in-place modified list.

    Examples:
        >>> messages = [Message(content="Hello")]
        >>> modified_messages = append_to_messages(messages, Message(content="New"))
        >>> modified_messages
        [Message(role='user', content='Hello'), Message(role='user', content='New')]
        >>> id(messages) == id(modified_messages)
        True
    """
    messages.append(new_message)
    return messages


def append_to_sys(user_content: str, sys_content: str | None = None) -> list[Message]:
    """Appends a user message to a list of messages, optionally including a system message.

    Args:
        user_content: The content of the user message.
        sys_content: Optional content for the system message. Defaults to None.

    Returns:
        A list of messages including the optional system message and the user message.

    Examples:
        >>> append_to_sys("Hello, world!")
        [Message(role='user', content='Hello, world!')]

        >>> append_to_sys("Hello, world!", "System initialized.")
        [Message(role='system', content='System initialized.'), Message(role='user', content='Hello, world!')]
    """
    sys = [Message(role="system", content=sys_content)] if sys_content else []
    return [*sys, Message(content=user_content)]


def prepend_sys(messages: Collection, sys_content: str) -> list[Message]:
    """Prepends a system message to a list of messages.

    Args:
        messages: The list of existing messages.
        sys_content: The content of the system message to be prepended.

    Returns:
        A new list of messages with the system message prepended.

    Examples:
        >>> messages = [Message(role="user", content="Hello!")]
        >>> prepend_sys(messages, "System initialized.")
        [Message(role='system', content='System initialized.'), Message(role='user', content='Hello!')]
    """
    return [Message(role="system", content=sys_content), *messages]


def indent_xml(xml_string, indent_size=2):
    output = []
    indent_level = 0

    # Split the input XML into parts by tags
    parts = xml_string.replace(">", ">\n").replace("<", "\n<").split("\n")
    parts = [part for part in parts if part.strip()]  # Remove empty parts

    for part in parts:
        if part.startswith("</"):
            # Closing tag, decrease indent
            indent_level -= indent_size
            output.append(" " * indent_level + part)
        elif part.startswith("<") and not part.endswith("/>") and ">" in part:
            # Opening tag, maintain then increase indent
            output.append(" " * indent_level + part)
            indent_level += indent_size
        elif part.endswith("/>"):
            # Self-closing tag, just append
            output.append(" " * indent_level + part)
        else:
            # Text or other data, maintain current indent
            # Handle multiple lines within text nodes
            text_lines = part.split("\n")
            output.extend([
                " " * indent_level + line.strip() for line in text_lines if line.strip()
            ])

    return "\n".join(output)


def prepend_sys_and_append_sys(
    messages: Iterable[Message], initial_sys_content: str, final_sys_content: str
) -> list[Message]:
    return [
        Message(role="system", content=initial_sys_content),
        *messages,
        Message(role="system", content=final_sys_content),
    ]



import logging
import logging.config
import os
from typing import Any

import litellm


def configure_log_levels() -> None:
    """Configure log levels."""
    # Set sane default LiteLLM logging configuration
    # SEE: https://docs.litellm.ai/docs/observability/telemetry
    litellm.telemetry = False
    if (
        logging.getLevelNamesMapping().get(
            os.environ.get("LITELLM_LOG", ""), logging.WARNING
        )
        < logging.WARNING
    ):
        # If LITELLM_LOG is DEBUG or INFO, don't change the LiteLLM log levels
        litellm_loggers_config: dict[str, Any] = {}
    else:
        litellm_loggers_config = {
            "LiteLLM": {"level": "WARNING"},
            "LiteLLM Proxy": {"level": "WARNING"},
            "LiteLLM Router": {"level": "WARNING"},
        }

    logging.config.dictConfig({
        "version": 1,
        "disable_existing_loggers": False,
        # Lower level for httpx and LiteLLM
        "loggers": {"httpx": {"level": "WARNING"}} | litellm_loggers_config,
    })


def configure_stdout_logs(
    level: int | str = logging.INFO,
    fmt: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
) -> None:
    """Configure root logger to log to stdout.

    Args:
        level: Log level to be emitted to stdout.
        fmt: Optional format string.
    """
    logging.config.dictConfig({
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {"standard": {"format": fmt}},
        "handlers": {
            "stdout": {
                "level": "INFO",
                "formatter": "standard",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
            },
        },
        "root": {"level": level, "handlers": ["stdout"]},
    })


def discounted_returns(
    rewards: list[float], terminated: list[bool], discount: float = 1.0
) -> list[float]:
    r"""
    Calculate the discounted returns for a list of rewards, considering termination flags and a discount factor.

    The discounted return represents the future discounted rewards from each time step onwards, taking into account
    whether an episode has terminated at each step.

    The discounted return \( G_t \) is given by:

    .. math::
        G_t = \sum_{k=1}^{\infty} \gamma^{k-1} R_{t+k}

        where:
        - \( G_t \) is the discounted return starting from time step \( t \).
        - \( \gamma \) is the discount factor.
        - \( R_{t+k} \) is the reward received at time step \( t+k \).

    NOTE: this could live in ldp.alg, but it's here to avoid circular imports.

    Args:
        rewards: A list of rewards at each time step.
        terminated: A list of boolean flags indicating whether the episode terminated at each time step.
        discount: Discount factor to apply to future rewards. Defaults to 1.0 which means no discounting is applied.

    Returns:
        A list of discounted returns (rewards to go), with each element representing the
            total discounted reward from that step onwards.

    Example:
        >>> rewards = [1.0, 2.0, 3.0]
        >>> terminated = [False, False, True]
        >>> discounted_returns(rewards, terminated, discount=0.9)
        [5.23, 4.7, 3.0]
    """
    returns = []
    r = 0.0
    for reward, term in zip(reversed(rewards), reversed(terminated), strict=False):
        # 1 - term is 0 if the episode has terminated
        r = reward + discount * r * (1 - term)
        returns.append(r)
    returns.reverse()
    return returns



import importlib
import os
from unittest.mock import patch

import ldp.shims


def test_tqdm_import() -> None:
    assert ldp.shims.tqdm.__module__ == "tqdm.std"
    with patch.dict(os.environ, {"LDP_TQDM_USE_RICH": "1"}):
        importlib.reload(ldp.shims)
        assert ldp.shims.tqdm.__module__ == "tqdm.rich"



from __future__ import annotations

import logging
from collections.abc import Iterable
from typing import ClassVar, Protocol, Self, cast, runtime_checkable

from pydantic import BaseModel, ConfigDict, Field, JsonValue

from ldp.agent import MemoryAgent
from ldp.alg.optimizer.opt import Optimizer
from ldp.data_structures import Trajectory
from ldp.graph import CallID, Memory, MemoryOp, Op, OpResult
from ldp.graph.ops import TOutput

logger = logging.getLogger(__name__)


@runtime_checkable
class MemoryFactory(Protocol):
    async def __call__(
        self,
        memory_op: MemoryOp,
        output_op: Op[TOutput],
        memory_template: str,
        example_buffer: Iterable[tuple[CallID, CallID, float, JsonValue]],
    ) -> list[Memory]:
        """
        Create new memories from the example buffer.

        Args:
            memory_op: MemoryOp whose context contains the MemoryModel's query and input.
            output_op: Op whose context contains an output that can be correlated with
                how good the outcome was.
            memory_template: Template used for the Memory's string representation.
            example_buffer: Buffer of 4-tuples containing the memory_op's call ID, the
                output_op's call ID, the current discounted return, and arbitrary JSON
                metadata (which can be used to add task-specific metadata the memory).

        Returns:
            New Memories created.
        """


async def _default_memory_factory(
    memory_op: MemoryOp,
    output_op: Op[TOutput],
    memory_template: str,
    example_buffer: Iterable[tuple[CallID, CallID, float, JsonValue]],
) -> list[Memory]:
    return [
        Memory.from_ops(
            memory_op,
            mem_call_id,
            output_op,
            output_call_id,
            d_return,
            template=memory_template,
            metadata=metadata,
        )
        for mem_call_id, output_call_id, d_return, metadata in example_buffer
    ]


class MemoryOpt(BaseModel, Optimizer):
    """Trainer for memory agents. By default it is a minimizer.

    This optimizer simply adds memories to the MemoryOp using a memory factory.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    # Working around https://github.com/pydantic/pydantic/issues/10551
    default_memory_factory: ClassVar[MemoryFactory] = _default_memory_factory

    ### Configuration
    memory_op: MemoryOp
    output_op: Op
    reward_discount: float = 1.0
    memory_factory: MemoryFactory = Field(
        default=default_memory_factory,
        description=(
            "Async function to make Memories from an example buffer. It's async so this"
            " can involve an LLM completion if desired."
        ),
        exclude=True,
    )
    memory_template: str = Field(
        default="Input: {input}\nOutput: {output}\nReward: {value}",
        description="Template for a Memory's string representation.",
    )

    ### State
    steps: int = 0
    example_buffer: list[tuple[CallID, CallID, float, JsonValue]] = Field(
        default_factory=list
    )

    @classmethod
    def from_agent(cls, agent: MemoryAgent, **kwargs) -> Self:
        return cls(memory_op=agent._memory_op, output_op=agent._llm_call_op, **kwargs)

    def _memory_filter(
        self, call_id: CallID, memory_op: MemoryOp, d_return: float
    ) -> bool:
        # only keep memories that backprop reached, i.e. those that were used in
        # achieving the reward
        return memory_op.ctx.get(call_id, "grad_output", default=None) is not None

    def aggregate_trajectory(self, trajectory: Trajectory) -> None:
        if trajectory.failed:
            return

        d_returns = trajectory.compute_discounted_returns(self.reward_discount)

        for step, d_return in zip(trajectory.steps, d_returns, strict=True):
            output_run_id = cast(OpResult, step.action).call_id.run_id
            output_call_ids = self.output_op.get_call_ids({output_run_id})

            for output_call_id in output_call_ids:
                output = cast(
                    OpResult, self.output_op.ctx.get(output_call_id, "output")
                )
                mem_call_ids = {
                    m.call_id
                    for m in output.get_upstream_results(self.memory_op)
                    if self._memory_filter(m.call_id, self.memory_op, d_return)
                }

                self.example_buffer.extend(
                    (
                        mem_call_id,
                        output_call_id,
                        d_return,
                        {
                            "timestep": step.timestep,
                            "done": step.done,
                            "truncated": step.truncated,
                        },
                    )
                    for mem_call_id in mem_call_ids
                )

    async def update(self) -> None:
        """Create new memories from the example buffer and add them to MemoryOp."""
        for memory in await self.memory_factory(  # pylint: disable=too-many-function-args
            self.memory_op, self.output_op, self.memory_template, self.example_buffer
        ):
            await self.memory_op.memory_model.add_memory(memory)
        self.steps += 1
        self.example_buffer.clear()


class PositiveMemoryOpt(MemoryOpt):
    def _memory_filter(
        self, call_id: CallID, memory_op: MemoryOp, d_return: float
    ) -> bool:
        # only keep positive memories
        return d_return > 0 and super()._memory_filter(call_id, memory_op, d_return)



import asyncio
import json
from collections.abc import AsyncGenerator, Callable, Iterable
from datetime import datetime
from typing import Any, ClassVar, Self, cast
from uuid import UUID, uuid4

import litellm
from aviary.message import Message
from aviary.tools import Tool, ToolRequestMessage, ToolsAdapter
from aviary.utils import is_coroutine_callable
from pydantic import BaseModel, ConfigDict, Field, ValidationError, model_validator


class JSONSchemaValidationError(ValueError):
    """Raised when the completion does not match the specified schema."""


class LLMResult(BaseModel):
    """A class to hold the result of a LLM completion."""

    id: UUID = Field(default_factory=uuid4)
    config: dict | None = None
    prompt: list[Message] | None = Field(
        default=None, description="Messages sent to the LLM."
    )
    messages: list[Message] | None = Field(
        default=None, description="Messages received from the LLM."
    )
    prompt_count: int = Field(default=0, description="Count of prompt tokens.")
    completion_count: int = Field(default=0, description="Count of completion tokens.")
    model: str
    date: str = Field(default_factory=datetime.now().isoformat)
    seconds_to_first_token: float | None = None
    seconds_to_last_token: float = 0
    logprob: float | None = Field(
        default=None, description="Sum of logprobs in the completion."
    )
    system_fingerprint: str | None = Field(
        default=None, description="System fingerprint received from the LLM."
    )

    @property
    def prompt_and_completion_costs(self) -> tuple[float, float]:
        """Get a two-tuple of prompt tokens cost and completion tokens cost, in USD."""
        return litellm.cost_per_token(
            self.model,
            prompt_tokens=self.prompt_count,
            completion_tokens=self.completion_count,
        )

    @property
    def provider(self) -> str:
        """Get the model provider's name (e.g. "openai", "mistral")."""
        return litellm.get_llm_provider(self.model)[1]

    def get_supported_openai_params(self) -> list[str] | None:
        """Get the supported OpenAI parameters for the model."""
        return litellm.get_supported_openai_params(self.model)


def sum_logprobs(choice: litellm.utils.Choices) -> float | None:
    """Calculate the sum of the log probabilities of an LLM completion (a Choices object).

    Args:
        choice: A sequence of choices from the completion.

    Returns:
        The sum of the log probabilities of the choice.
    """
    try:
        logprob_obj = choice.logprobs
    except AttributeError:
        return None
    if isinstance(logprob_obj, dict):
        if logprob_obj.get("content"):
            return sum(
                logprob_info["logprob"] for logprob_info in logprob_obj["content"]
            )
    elif choice.logprobs.content:
        return sum(logprob_info.logprob for logprob_info in choice.logprobs.content)
    return None


def validate_json_completion(
    completion: litellm.ModelResponse, output_type: type[BaseModel]
) -> None:
    """Validate a completion against a JSON schema.

    Args:
        completion: The completion to validate.
        output_type: The Pydantic model to validate the completion against.
    """
    try:
        for choice in completion.choices:
            output_type.model_validate_json(choice.message.content or "")  # type: ignore[union-attr]
    except ValidationError as err:
        raise JSONSchemaValidationError(
            "The completion does not match the specified schema."
        ) from err


class MultipleCompletionLLMModel(BaseModel):
    """Run n completions at once, all starting from the same messages."""

    model_config = ConfigDict(extra="forbid")

    # this should keep the original model
    # if fine-tuned, this should still refer to the base model
    name: str = "unknown"
    config: dict = Field(
        default={
            "model": "gpt-3.5-turbo",  # Default model should have cheap input/output for testing
            "temperature": 0.1,
        }
    )
    encoding: Any | None = None

    def __str__(self) -> str:
        return f"{type(self).__name__} {self.name}"

    @model_validator(mode="after")
    def set_model_name(self) -> Self:
        if (
            self.config.get("model") in {"gpt-3.5-turbo", None}
            and self.name != "unknown"
            or self.name != "unknown"
            and "model" not in self.config
        ):
            self.config["model"] = self.name
        elif "model" in self.config and self.name == "unknown":
            self.name = self.config["model"]
        # note we do not consider case where both are set
        # because that could be true if the model is fine-tuned
        return self

    async def achat(
        self, messages: Iterable[Message], **kwargs
    ) -> litellm.ModelResponse:
        return await litellm.acompletion(
            messages=[m.model_dump(by_alias=True) for m in messages],
            **(self.config | kwargs),
        )

    async def achat_iter(self, messages: Iterable[Message], **kwargs) -> AsyncGenerator:
        return cast(
            AsyncGenerator,
            await litellm.acompletion(
                messages=[m.model_dump(by_alias=True) for m in messages],
                stream=True,
                stream_options={
                    "include_usage": True,  # Included to get prompt token counts
                },
                **(self.config | kwargs),
            ),
        )

    # SEE: https://platform.openai.com/docs/api-reference/chat/create#chat-create-tool_choice
    # > `required` means the model must call one or more tools.
    TOOL_CHOICE_REQUIRED: ClassVar[str] = "required"

    async def call(  # noqa: C901, PLR0915
        self,
        messages: list[Message],
        callbacks: list[Callable] | None = None,
        output_type: type[BaseModel] | None = None,
        tools: list[Tool] | None = None,
        tool_choice: Tool | str | None = TOOL_CHOICE_REQUIRED,
        **chat_kwargs,
    ) -> list[LLMResult]:
        start_clock = asyncio.get_running_loop().time()

        # Deal with tools. OpenAI throws an error if tool list is empty,
        # so skip this block if tools in (None, [])
        if tools:
            chat_kwargs["tools"] = ToolsAdapter.dump_python(
                tools, exclude_none=True, by_alias=True
            )
            if tool_choice is not None:
                chat_kwargs["tool_choice"] = (
                    {
                        "type": "function",
                        "function": {"name": tool_choice.info.name},
                    }
                    if isinstance(tool_choice, Tool)
                    else tool_choice
                )

        # deal with specifying output type
        if output_type is not None:
            schema = json.dumps(
                output_type.model_json_schema(mode="serialization"), indent=2
            )
            schema_msg = f"Respond following this JSON schema:\n\n{schema}"
            # Get the system prompt and its index, or the index to add it
            i, system_prompt = next(
                ((i, m) for i, m in enumerate(messages) if m.role == "system"),
                (0, None),
            )
            messages = [
                *messages[:i],
                system_prompt.append_text(schema_msg, inplace=False)
                if system_prompt
                else Message(role="system", content=schema_msg),
                *messages[i + 1 if system_prompt else i :],
            ]
            chat_kwargs["response_format"] = {"type": "json_object"}

        # add static configuration to kwargs
        chat_kwargs = self.config | chat_kwargs
        n = chat_kwargs.get("n", 1)  # number of completions
        if n < 1:
            raise ValueError("Number of completions (n) must be >= 1.")

        prompt = [
            m
            if not isinstance(m, ToolRequestMessage) or m.tool_calls
            # OpenAI doesn't allow for empty tool_calls lists, so downcast empty
            # ToolRequestMessage to Message here
            else Message(role=m.role, content=m.content)
            for m in messages
        ]
        results: list[LLMResult] = []

        if callbacks is None:
            completion: litellm.ModelResponse = await self.achat(prompt, **chat_kwargs)
            if output_type is not None:
                validate_json_completion(completion, output_type)

            for choice in completion.choices:
                if isinstance(choice, litellm.utils.StreamingChoices):
                    raise NotImplementedError("Streaming is not yet supported.")

                if (
                    tools is not None  # Allows for empty tools list
                    or choice.finish_reason == "tool_calls"
                    or (getattr(choice.message, "tool_calls", None) is not None)
                ):
                    serialized_choice_message = choice.message.model_dump()
                    serialized_choice_message["tool_calls"] = (
                        serialized_choice_message.get("tool_calls") or []
                    )
                    output_messages: list[Message | ToolRequestMessage] = [
                        ToolRequestMessage(**serialized_choice_message)
                    ]
                else:
                    output_messages = [Message(**choice.message.model_dump())]

                results.append(
                    LLMResult(
                        model=self.name,
                        config=chat_kwargs,
                        prompt=prompt,
                        messages=output_messages,
                        logprob=sum_logprobs(choice),
                        system_fingerprint=completion.system_fingerprint,
                        # Note that these counts are aggregated over all choices
                        completion_count=completion.usage.completion_tokens,  # type: ignore[attr-defined,unused-ignore]
                        prompt_count=completion.usage.prompt_tokens,  # type: ignore[attr-defined,unused-ignore]
                    )
                )
        else:
            if tools:
                raise NotImplementedError("Using tools with callbacks is not supported")
            if n > 1:
                raise NotImplementedError(
                    "Multiple completions with callbacks is not supported"
                )
            result = LLMResult(model=self.name, config=chat_kwargs, prompt=prompt)

            sync_callbacks = [f for f in callbacks if not is_coroutine_callable(f)]
            async_callbacks = [f for f in callbacks if is_coroutine_callable(f)]
            stream_completion = await self.achat_iter(messages, **chat_kwargs)
            text_result = []
            role = "assistant"

            async for chunk in stream_completion:
                delta = chunk.choices[0].delta
                role = delta.role or role
                if delta.content:
                    s = delta.content
                    if result.seconds_to_first_token == 0:
                        result.seconds_to_first_token = (
                            asyncio.get_running_loop().time() - start_clock
                        )
                    text_result.append(s)
                    [await f(s) for f in async_callbacks]
                    [f(s) for f in sync_callbacks]
                if hasattr(chunk, "usage"):
                    result.prompt_count = chunk.usage.prompt_tokens

            output = "".join(text_result)
            result.completion_count = litellm.token_counter(
                model=self.name,
                text=output,
            )
            # TODO: figure out how tools stream, and log probs
            result.messages = [Message(role=role, content=output)]
            results.append(result)

        if not results:
            # This happens in unit tests. We should probably not keep this block around
            # long-term. Previously, we would emit an empty ToolRequestMessage if
            # completion.choices were empty, so  I am replicating that here.
            results.append(
                LLMResult(
                    model=self.name,
                    config=chat_kwargs,
                    prompt=prompt,
                    messages=[ToolRequestMessage(tool_calls=[])],
                )
            )

        end_clock = asyncio.get_running_loop().time()

        for result in results:
            # Manually update prompt count if not set, which can
            # happen if the target model doesn't support 'include_usage'
            if not result.prompt_count:
                result.prompt_count = litellm.token_counter(
                    model=self.name,
                    messages=[m.model_dump() for m in result.messages],  # type: ignore[union-attr]
                )

            # update with server-side counts
            result.seconds_to_last_token = end_clock - start_clock

        return results


class LLMModel(MultipleCompletionLLMModel):
    async def call(self, *args, **kwargs) -> LLMResult:  # type: ignore[override]
        return (await super().call(*args, **kwargs))[0]



"""This module contains loss Op implementations."""

from typing import TYPE_CHECKING

import tree

from .op_utils import CallID
from .ops import GradInType, Op, OpCtx

if TYPE_CHECKING:
    import numpy.typing as npt
    import torch


class MSELossOp(Op):
    async def forward(
        self,
        prediction: "npt.NDArray | torch.Tensor",
        target: "npt.NDArray | torch.Tensor",
    ) -> "float | torch.Tensor":
        return ((prediction - target) ** 2).mean()

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args,
        input_kwargs,
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        prediction = input_kwargs["prediction"]
        target = input_kwargs["target"]
        grad = 2 * (prediction - target)
        return [], {"prediction": grad, "target": None}



"""This module contains commonly-used Op implementations."""

from __future__ import annotations

import asyncio
import functools
import inspect
import logging
from collections.abc import Awaitable, Callable
from functools import lru_cache
from typing import Generic, TypeVar, cast, overload

import numpy as np
import tree
from aviary.message import Message
from aviary.tools import Tool, ToolRequestMessage
from aviary.utils import is_coroutine_callable
from pydantic import BaseModel

from ldp.llms import (
    EmbeddingModel,
    HybridEmbeddingModel,
    LiteEmbeddingModel,
    LLMModel,
    LLMResult,
    SparseEmbeddingModel,
)

from .gradient_estimators import assign_constant_grads
from .memory import Memory, MemoryModel, UIndexMemoryModel
from .op_utils import CallID, get_call_id, get_training_mode
from .ops import GradInType, Op, OpCtx, ResultOrValue, TOutput

logger = logging.getLogger(__name__)


def logsumexp(a: np.ndarray | list[float]) -> float:
    a_max = np.max(a)
    return a_max + np.log(np.sum(np.exp(a - a_max)))


class IdentityOp(Op[TOutput]):
    """
    An operation that simply returns the input value.

    NOTE: this op is equivalent to FxnOp(lambda x: x).
    """

    async def forward(self, value: TOutput) -> TOutput:
        # We assume value already has the correct run_id from its producer
        return value

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        return [], {"value": grad_output}


class StopGradOp(IdentityOp[TOutput]):
    """Pass through Op that terminates gradients in the backward pass."""

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        return assign_constant_grads(input_args, input_kwargs, None)


TConfig = TypeVar("TConfig", bound=BaseModel | dict)


class ConfigOp(Op[TConfig], Generic[TConfig]):
    """An operation that contains a configuration object."""

    def __init__(self, config: TConfig):
        self.config = config

    async def forward(self) -> TConfig:
        return self.config

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        # Check that the grad_output structure is consistent with our config
        tree.assert_same_structure(
            grad_output, ctx.get(call_id, "output").value, check_types=False
        )

        # Terminate here - we're a leaf since a ConfigOp takes no inputs
        return [], {}


TResult = TypeVar("TResult")


class Cacheable(Generic[TResult]):
    def __init__(self, co: Awaitable[TResult]) -> None:
        self.co = co
        self.done = False
        self.result: TResult | None = None
        self.lock = asyncio.Lock()

    async def get_result(self) -> TResult | None:
        async with self.lock:
            if not self.done:
                self.result = await self.co
                self.done = True
            return self.result

    def __await__(self):
        return self.get_result().__await__()


def async_cache(func):
    @functools.lru_cache(maxsize=1024)
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return Cacheable(co=func(*args, **kwargs))

    return wrapper


class FxnOp(Op[TOutput]):
    """
    Wrap a function for a straight through gradient approximation for all args/kwargs.

    Basically, consider the fxn as a transform upon the inputs during the forward pass,
    and propagating the same gradient for all inputs during the backward pass.
    """

    def __init__(
        self,
        fxn: Callable[..., TOutput] | Callable[..., Awaitable[TOutput]],
        cache: bool = False,
        fxn_name: str | None = None,  # useful for lambdas
    ):
        if cache:
            self.fxn = (
                async_cache(fxn) if is_coroutine_callable(fxn) else lru_cache()(fxn)
            )
        else:
            self.fxn = fxn

        try:
            self.fxn_name = fxn_name or fxn.__name__
        except AttributeError:  # unittest.mock.Mock or lambda
            self.fxn_name = str(fxn)

        # override forward args with the signature of the function
        fwd_sig = inspect.signature(self.fxn)
        self._fwd_args = list(fwd_sig.parameters.values())

    def __repr__(self) -> str:
        return f"{type(self).__name__} {self.fxn_name} ({id(self)})"

    async def forward(self, *args, **kwargs) -> TOutput:
        if is_coroutine_callable(self.fxn):
            return await self.fxn(*args, **kwargs)
        return self.fxn(*args, **kwargs)

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        return assign_constant_grads(input_args, input_kwargs, 0.0)


class PromptOp(FxnOp[str]):
    """An operation that formats kwargs into a prompt string."""

    async def _fxn(
        self, prompt_kwargs: dict[str, str] | None = None, **kwargs: str
    ) -> str:
        return self.prompt.format(**{**(prompt_kwargs or {}), **kwargs})

    def __init__(self, prompt: str):
        self.prompt = prompt
        super().__init__(fxn=self._fxn, cache=False)

    def __repr__(self) -> str:
        # we want to use Op.__repr__, not FxnOp.__repr__
        return super(FxnOp, self).__repr__()


class LLMCallOp(Op[Message]):
    """An operation for LLM calls interaction."""

    def __init__(self, num_samples_logprob_estimate: int = 0) -> None:
        super().__init__()
        self.num_samples_partition_estimate = num_samples_logprob_estimate

    @overload
    async def forward(
        self,
        config: dict,
        msgs: list[Message],
        tools: list[Tool] = ...,
        tool_choice: Tool | str | None = LLMModel.TOOL_CHOICE_REQUIRED,
    ) -> ToolRequestMessage: ...

    @overload
    async def forward(
        self,
        config: dict,
        msgs: list[Message],
        tools: None = None,
        tool_choice: str | None = LLMModel.TOOL_CHOICE_REQUIRED,
    ) -> Message: ...

    async def forward(
        self,
        config: dict,
        msgs: list[Message],
        tools: list[Tool] | None = None,
        tool_choice: Tool | str | None = LLMModel.TOOL_CHOICE_REQUIRED,
    ) -> Message:
        model = LLMModel(config=config)

        result = await model.call(messages=msgs, tools=tools, tool_choice=tool_choice)
        if result.messages is None:
            raise ValueError("No messages returned")

        # if not set, assume temp = 1. TODO: when would it not be set?
        temperature: float = (result.config or {}).get("temperature", 1.0)

        # Compute a Monte Carlo estimate of the logprob of this sequence at the given temperature.
        logprob = await self.compute_logprob(
            raw_log_p=result.logprob,
            temperature=temperature,
            model=model,
            messages=msgs,
            tools=tools,
            tool_choice=tool_choice,
        )

        call_id = get_call_id()
        self.ctx.update(call_id, "result", result)
        # This is the logprob of this sequence according to the raw model, without
        # any temperature/top-p distribution shaping.
        self.ctx.update(call_id, "raw_logprob", result.logprob)

        self.ctx.update(call_id, "temperature", temperature)
        self.ctx.update(call_id, "logprob", logprob)

        return result.messages[0]

    async def compute_logprob(
        self,
        raw_log_p: float | None,
        temperature: float,
        model: LLMModel,
        **model_kwargs,
    ) -> float | None:
        """This method computes a Monte Carlo estimate of logprob for a given temperature.

        It takes as input the logprob at T=1. The derivation is in Section 5.1 of the Aviary notes.
        """
        if temperature == 1:
            return raw_log_p

        if temperature == 0:
            return 1.0

        if raw_log_p is None or self.num_samples_partition_estimate == 0:
            return None

        # TODO: Try using n completions from a single API call. Need to modify LLMModel.call to do this, since
        # it currently only checks completion.choices[0]. Would reduce cost for long prompts.
        # TODO: think about whether sampling params besides temperature need to be accounted for, like top_p
        results = await asyncio.gather(*[
            model.call(temperature=1, **model_kwargs)
            for _ in range(self.num_samples_partition_estimate)
        ])
        temp_factor = 1.0 / temperature - 1.0

        # Partition function estimate:
        # Z_T = E_P[ e^(lnP/T - lnP) ]
        log_Z_T = logsumexp([
            temp_factor * cast(float, result.logprob) for result in results
        ]) - np.log(self.num_samples_partition_estimate)

        return (raw_log_p / temperature) - log_Z_T

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        # By default, we want to descend into config, but not msgs/tools/tool_choice
        # Essentially: we can think of each config field as an independent parameter,
        # but not necessarily each message or tool.

        # tree.map_structure allows us to assign a gradient of 0 to all fields of config
        grad_config = tree.map_structure(lambda _: 0.0, input_kwargs["config"])
        grad_kwargs = {"config": grad_config}
        for arg in ("msgs", "tools", "tool_choice"):
            if arg in input_kwargs:
                grad_kwargs[arg] = 0.0

        return [], grad_kwargs

    def get_examples(self) -> list[tuple[LLMResult, float]]:
        examples = [
            (
                self.ctx.get(c, "result", None),
                # get 'model' kwarg from grad_input
                # use default of None if not found
                self.ctx.get(c, "grad_input", default=([], {}))[1].get("model", None),
            )
            for c in self.get_call_ids()
        ]
        # filter out the None values
        return [(e, w) for e, w in examples if e is not None]


class MemoryOp(Op[list[Memory]]):
    """An operation for managing memory retrieval and storage."""

    def __init__(self, memory_model: MemoryModel | None = None):
        super().__init__()
        self.memory_model = memory_model or UIndexMemoryModel(
            embedding_model=EmbeddingModel.from_name("sparse")
        )

    async def forward(
        self,
        query: str,
        input: str | None = None,  # noqa: A002
        matches: int = 3,
    ) -> list[Memory]:
        """Retrieve relevant memories based on a query."""
        if get_training_mode():
            call_id = get_call_id()
            self.ctx.update(call_id, "query", query)
            self.ctx.update(call_id, "memory_input", input)
        return await self.memory_model.get_memory(query, matches)

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        """Backward pass for memory retrieval - goes back to item."""
        return assign_constant_grads(input_args, input_kwargs, 0.0)


class EmbeddingOp(Op):
    """A general operation for embedding text using LiteLLM."""

    def __init__(
        self,
        *,
        dense_embedding: str = "text-embedding-3-small",
        dense_embedding_dim: int = 512,
        sparse_embedding_dim: int = 0,
        **embedding_model_kwargs,
    ):
        if "timeout" not in embedding_model_kwargs:
            embedding_model_kwargs.setdefault("timeout", 60)
        emb_models: list[EmbeddingModel] = []
        if dense_embedding_dim > 0:
            emb_models.append(
                LiteEmbeddingModel(
                    name=dense_embedding,
                    dimensions=dense_embedding_dim,
                    embed_kwargs=embedding_model_kwargs,
                )
            )
        if sparse_embedding_dim > 0:
            emb_models.append(SparseEmbeddingModel(dimensions=sparse_embedding_dim))
        self.embedding = HybridEmbeddingModel(models=emb_models)

    async def forward(self, string_input: str) -> np.ndarray:
        return await self.embedding.embed_text(string_input)

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args,
        input_kwargs,
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        return [], {"string_input": None}



import pytest

from ldp.alg.optimizer.replay_buffers import CircularReplayBuffer


def test_circular_buffer():
    buf = CircularReplayBuffer()

    samples = [{"state": 1, "action": 2, "reward": 3, "t": t} for t in range(5)]
    buf += samples
    buf.resize(3)  # should eject t=0, 1
    assert {sample["t"] for sample in buf} == {2, 3, 4}

    # check we can iterate
    next(buf.batched_iter(batch_size=3))

    # add a bad sample
    buf.append({})
    with pytest.raises(
        RuntimeError, match="Found buffer element with inconsistent keys"
    ):
        next(buf.batched_iter(batch_size=4))



from typing import Any

from aviary.message import Message
from pydantic import BaseModel, Field

from ldp.graph import ConfigOp, FxnOp, LLMCallOp, PromptOp, compute_graph
from ldp.graph.ops import ResultOrValue
from ldp.llms import append_to_sys
from ldp.llms.prompts import indent_xml


class ReflectModuleConfig(BaseModel):
    """Configuration for the ReflectModuleConfig."""

    llm_model: dict[str, Any] = Field(
        default={"model": "gpt-3.5-turbo"},
        description="Starting configuration for the LLM model.",
    )


class ReflectModule:
    """A module that simply gives an LLM to reflect on an input."""

    def __init__(self, start_config: ReflectModuleConfig):
        self.llm_call_op = LLMCallOp()
        self.prompt_op = PromptOp(
            "Consider a proposed response based on context. Reflect on the response"
            " within <thought> tags then conclude with a possibly revised response"
            " within <final-response> tags."
        )
        self.config_op = ConfigOp[ReflectModuleConfig](config=start_config)
        self.llm_config_op = FxnOp[dict](lambda c: c.llm_model)
        self.package_fxn = FxnOp(append_to_sys)

        def extract_msg(msg: Message, backup_response: str) -> str:
            msg_str = msg.content
            if msg_str and "<final-response>" in msg_str:
                return msg_str.split("<final-response>")[1].split("</final-response>")[
                    0
                ]
            if msg_str and "<response>" in msg_str:
                return msg_str.split("<response>")[1].split("</response>")[0]
            return backup_response

        self.extract_msg = FxnOp(extract_msg)

    @compute_graph()
    async def __call__(
        self, context: ResultOrValue[str], response: ResultOrValue[str]
    ) -> ResultOrValue[str]:
        llm_config = await self.llm_config_op(await self.config_op())
        sys_str = await self.prompt_op()
        user_str = indent_xml(
            f"<context>{context}</context><response>{response}</response>"
        )
        msg = await self.package_fxn(user_str, sys_str)
        llm_result = await self.llm_call_op(llm_config, msg)
        return await self.extract_msg(llm_result, response)
